import pytest
from pathlib import Path
from xarray import Dataset
from collections import namedtuple
import dask
import dask.array as da
from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
pmp = pytest.mark.parametrize


def test_model2comps(ms_name):
    '''
    TODO - This only tests the separate function implementations
    not the workers. Tested by spotless workflow?
    '''

    import numpy as np
    np.random.seed(420)
    from numpy.testing import assert_allclose
    from daskms import xds_from_ms, xds_from_table, xds_to_table
    from daskms.experimental.zarr import xds_to_zarr
    from pfb.utils.misc import Gaussian2D, give_edges
    from africanus.constants import c as lightspeed
    from pfb.utils.misc import (fit_image_cube,
                                eval_coeffs_to_cube,
                                eval_coeffs_to_slice)

    test_dir = Path(ms_name).resolve().parent
    xds = xds_from_ms(ms_name,
                      chunks={'row': -1, 'chan': -1})[0]
    spw = xds_from_table(f'{ms_name}::SPECTRAL_WINDOW')[0]

    utime = np.unique(xds.TIME.values)
    freq = spw.CHAN_FREQ.values.squeeze()
    freq0 = np.mean(freq)

    ntime = utime.size
    nchan = freq.size
    nant = np.maximum(xds.ANTENNA1.values.max(), xds.ANTENNA1.values.max()) + 1

    ncorr = xds.corr.size

    uvw = xds.UVW.values
    nrow = uvw.shape[0]
    u_max = abs(uvw[:, 0]).max()
    v_max = abs(uvw[:, 1]).max()
    uv_max = np.maximum(u_max, v_max)

    # image size
    cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

    srf = 2.0
    cell_rad = cell_N / srf
    cell_deg = cell_rad * 180 / np.pi
    cell_size = cell_deg * 3600
    print("Cell size set to %5.5e arcseconds" % cell_size)

    from ducc0.fft import good_size
    # the test will fail in intrinsic if sources fall near beam sidelobes
    fov = 1.0
    npix = good_size(int(fov / cell_deg))
    while npix % 2:
        npix += 1
        npix = good_size(npix)

    nx = npix
    ny = npix

    print("Image size set to (%i, %i, %i)" % (nchan, nx, ny))

    # model
    model = np.zeros((nchan, nx, ny), dtype=np.float64)
    nsource = 25
    border = np.maximum(int(0.15*nx), int(0.15*ny))
    Ix = np.random.randint(border, npix-border, nsource)
    Iy = np.random.randint(border, npix-border, nsource)
    alpha = -0.7 + 0.1 * np.random.randn(nsource)
    I0 = 1.0 + np.exp(np.random.randn(nsource))
    extentx = np.random.randint(3, int(0.1*nx), nsource)
    extenty = np.random.randint(3, int(0.1*nx), nsource)
    pas = np.random.random(nsource) * 180
    x = -(nx/2) + np.arange(nx)
    y = -(nx/2) + np.arange(ny)
    xin, yin = np.meshgrid(x, y, indexing='ij')
    for i in range(nsource):
        emaj = np.maximum(extentx[i], extenty[i])
        emin = np.minimum(extentx[i], extenty[i])
        gauss = Gaussian2D(xin, yin, GaussPar=(emaj, emin, pas[i]))
        mx, my, gx, gy = give_edges(Ix[i], Iy[i], nx, ny, nx, ny)
        spectrum = I0[i] * (freq/freq0) ** alpha[i]
        model[:, mx, my] += spectrum[:, None, None] * gauss[None, gx, gy]

    mfreqs = freq
    mtimes = utime[0:1]  # arbitrary for now



    coeffs, Ix, Iy, expr, params, tfunc, ffunc = \
        fit_image_cube(mtimes, mfreqs, model[None, :, :, :], nbasisf=nchan,
                       sigmasq=0.0, method='Legendre')
    image = eval_coeffs_to_cube(mtimes, mfreqs, nx, ny, coeffs, Ix, Iy,
                                expr, params, tfunc, ffunc)

    image = image[0]  # no time axis for now
    mask = model > 0
    assert_allclose(image[mask], model[mask], atol=1e-10)

    # test spatial interpolation
    # if we shift the center by an integer number of pixels
    # the pixel centers should stay the same
    xshift = 25
    x0 = cell_deg*xshift
    yshift = -10
    y0 = cell_deg*yshift
    for npix in [100, nx, 2*nx]:
        imout = eval_coeffs_to_slice(mtimes[0],
                                    mfreqs[0],
                                    coeffs,
                                    Ix, Iy,
                                    expr,
                                    params,
                                    tfunc,
                                    ffunc,
                                    nx, ny,
                                    cell_deg, cell_deg,
                                    0.0, 0.0,
                                    npix, npix,
                                    cell_deg, cell_deg,
                                    x0, y0)

        mx, my, gx, gy = give_edges(npix//2 - xshift,
                                    npix//2 - yshift, npix, npix, nx, ny)

        assert_allclose(1.0 + imout[mx, my], 1.0 + model[0, gx, gy])

# test_model2comps()



import numpy as np
from functools import partial
import dask.array as da
from distributed import wait
from uuid import uuid4
from ducc0.misc import make_noncritical
from pfb.utils.misc import norm_diff
import pyscilog
log = pyscilog.get_logger('PCG')


def cg(A,
       b,
       x0=None,
       tol=1e-5,
       maxit=500,
       verbosity=1,
       report_freq=10):

    if x0 is None:
        x = np.zeros(b.shape, dtype=b.dtype)
    else:
        x = x0.copy()

    # initial residual
    r = A(x) - b
    p = -r
    rnorm = np.vdot(r, r)
    rnorm0 = rnorm
    eps = rnorm
    k = 0
    while eps > tol and k < maxit:
        xp = x
        rp = r
        Ap = A(p)
        alpha = rnorm/np.vdot(p, Ap)
        x += alpha * p
        r += alpha * Ap
        rnorm_next = np.vdot(r, r)
        beta = rnorm_next/rnorm
        p = beta*p - r
        rnorm = rnorm_next
        eps = rnorm #/rnorm0

        k += 1

    if k >= maxit:
        print(f"Max iters reached eps = {eps}")

    return x


def pcg(A,
        b,
        x0=None,
        M=None,
        tol=1e-5,
        maxit=500,
        minit=100,
        verbosity=1,
        report_freq=10,
        backtrack=True,
        return_resid=False):

    if x0 is None:
        x0 = np.zeros(b.shape, dtype=b.dtype)

    if M is None:
        def M(x): return x

    r = A(x0) - b
    y = M(r)
    if not np.any(y):
        print(f"Initial residual is zero", file=log)
        return x0
    p = -y
    rnorm = np.vdot(r, y)
    if np.isnan(rnorm) or rnorm == 0.0:
        eps0 = 1.0
    else:
        eps0 = rnorm
    k = 0
    x = x0
    eps = 1.0
    stall_count = 0
    while (eps > tol or k < minit) and k < maxit and stall_count < 5:
        xp = x.copy()
        rp = r.copy()
        Ap = A(p)
        rnorm = np.vdot(r, y)
        alpha = rnorm / np.vdot(p, Ap)
        x = xp + alpha * p
        r = rp + alpha * Ap
        y = M(r)
        rnorm_next = np.vdot(r, y)
        while rnorm_next > rnorm and backtrack:  # TODO - better line search
            alpha *= 0.75
            x = xp + alpha * p
            r = rp + alpha * Ap
            y = M(r)
            rnorm_next = np.vdot(r, y)

        beta = rnorm_next / rnorm
        p = beta * p - y
        # if p is zero we should stop
        if not np.any(p):
            break
        rnorm = rnorm_next
        k += 1
        epsp = eps
        eps = norm_diff(x, xp)
        # np.linalg.norm(x - xp) / np.linalg.norm(x)
        # epsn = rnorm / eps0
        # eps = rnorm / eps0
        # eps = np.maximum(epsx, epsn)

        # if np.abs(epsp - eps) < 1e-3*tol:
        #     stall_count += 1

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} epsx = {eps:.3e}",
                  file=log)

    if k >= maxit:
        if verbosity:
            print(f"Max iters reached. eps = {eps:.3e}", file=log)
    elif stall_count >= 5:
        if verbosity:
            print(f"Stalled after {k} iterations with eps = {eps:.3e}", file=log)
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations", file=log)
    if not return_resid:
        return x
    else:
        return x, r


def cg_dct(A,
           b,
           x,
           tol=1e-5,
           maxit=500,
           verbosity=1,
           report_freq=10):
    '''
    A specialised version of the pcg that doesn't need x to live on a
    single grid. As a result x is nested a dictionary keyed on field
    name and then on time and freq. There is currently no support for
    overlapping fields i.e. fields need to be distinct. This simplified
    version does not yet support bactracking or pre-conditioning so it
    will probably only work if A is a very good approximation of Ax=b
    '''

    def residual(Ax, b, r, p):
        for field in Ax.keys():
            for i in Ax[field].keys():
                r[field][i] = Ax[field][i] - b[field][i]
                p[field][i] = -r[field][i]
        return r, p

    def vdot_dct(a, b):
        res = 0.0
        for field in a.keys():
            for i in a[field].keys():
                res += np.vdot(a[field][i], b[field][i])
        return res

    def pluseq_dct(a, b, alpha=1.0):
        # implement a += alpha * b
        for field in a.keys():
            for i in a[field].keys():
                a[field][i] += alpha * b[field][i]
        return a

    def pupdate(p, r, beta=1):
        # implement p = beta * p - r
        for field in p.keys():
            for i in p[field].keys():
                p[field][i] = beta * p[field][i] - r[field][i]
        return p

    def norm_dct(a, b=None):
        norm = 0.0
        for field in a.keys():
            for i in a[field].keys():
                if b is not None:
                    norm += np.linalg.norm(a[field][i] - b[field][i])
                else:
                    norm += np.linalg.norm(a[field][i])
        return norm

    # initial residual
    Ax = A(x)
    r = {}
    p = {}
    for field in Ax.keys():
        r[field] = {}
        p[field] = {}
        for i in Ax[field].keys():
            r[field][i] = Ax[field][i] - b[field][i]
            p[field][i] = -r[field][i]


    rnorm = vdot_dct(r, r)
    rnorm0 = rnorm
    eps = rnorm
    k = 0
    while eps > tol and k < maxit:
        xp = x
        rp = r
        Ap = A(p)
        pAp = vdot_dct(p, Ap)
        alpha = rnorm/pAp  #np.vdot(p, Ap)
        x = pluseq_dct(x, p, alpha=alpha)
        r = pluseq_dct(r, Ap, alpha=alpha)
        # x += alpha * p
        # r += alpha * Ap
        rnorm_next = vdot_dct(r, r)
        beta = rnorm_next/rnorm
        p = pupdate(p, r, beta=beta)
        # p = beta*p - r
        rnorm = rnorm_next
        eps = rnorm
        # eps = norm_dct(x, b=xp) / norm_dct(x)

        k += 1

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} eps = {eps:.3e}",
                  file=log)

    if k >= maxit:
        if verbosity:
            print(f"Max iters reached. eps = {eps:.3e}", file=log)
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations", file=log)
    return x, r


from pfb.operators.hessian import _hessian_psf_slice
def _pcg_psf_impl(psfhat,
                  b,
                  x0,
                  beam,
                  lastsize,
                  nthreads,
                  sigmainv,
                  tol=1e-5,
                  maxit=500,
                  minit=100,
                  verbosity=1,
                  report_freq=10,
                  backtrack=True):
    '''
    A specialised distributed version of pcg when the operator implements
    convolution with the psf (+ L2 regularisation by sigma**2)
    '''
    nband, nx, ny = b.shape
    _, nx_psf, nyo2 = psfhat.shape
    model = np.zeros((nband, nx, ny), dtype=b.dtype, order='C')
    # PCG preconditioner
    if sigmainv > 0:
        def M(x): return x / sigmainv
    else:
        M = None

    for k in range(nband):
        xpad = np.empty((nx_psf, lastsize), dtype=b.dtype, order='C')
        xpad = make_noncritical(xpad)
        xhat = np.empty((nx_psf, nyo2), dtype=psfhat.dtype, order='C')
        xhat = make_noncritical(xhat)
        xout = np.empty((nx, ny), dtype=b.dtype, order='C')
        xout = make_noncritical(xout)
        A = partial(_hessian_psf_slice,
                    xpad,
                    xhat,
                    xout,
                    psfhat[k],
                    beam[k],
                    lastsize,
                    nthreads=nthreads,
                    sigmainv=sigmainv)

        model[k] = pcg(A, b[k], x0[k],
                       M=M, tol=tol, maxit=maxit, minit=minit,
                       verbosity=verbosity, report_freq=report_freq,
                       backtrack=backtrack)

    return model

def _pcg_psf(psfhat,
             b,
             x0,
             beam,
             lastsize,
             nthreads,
             sigmainv,
             cgopts):
    return _pcg_psf_impl(psfhat,
                         b,
                         x0,
                         beam,
                         lastsize,
                         nthreads,
                         sigmainv,
                         **cgopts)

def pcg_psf(psfhat,
            b,
            x0,
            beam,
            lastsize,
            nthreads,
            sigmainv,
            cgopts,
            compute=True):

    if not isinstance(x0, da.Array):
        x0 = da.from_array(x0, chunks=(1, -1, -1),
                          name="x-" + uuid4().hex)

    if not isinstance(psfhat, da.Array):
        psfhat = da.from_array(psfhat, chunks=(1, -1, -1),
                               name="psfhat-" + uuid4().hex)
    if not isinstance(b, da.Array):
        b = da.from_array(b, chunks=(1, -1, -1),
                          name="psfhat-" + uuid4().hex)

    if beam is None:
        bout = None
    else:
        bout = ('nb', 'nx', 'ny')
        if beam.ndim == 2:
            beam = beam[None]

        if not isinstance(beam, da.Array):
            beam = da.from_array(beam, chunks=(1, -1, -1),
                                 name="beam-" + uuid4().hex)
        if beam.shape[0] == 1:
            beam = da.tile(beam, (psfhat.shape[0], 1, 1))
        elif beam.shape[0] != psfhat.shape[0]:
            raise ValueError('Beam has incorrect shape')

    model = da.blockwise(_pcg_psf, ('nb', 'nx', 'ny'),
                         psfhat, ('nb', 'nx', 'ny'),
                         b, ('nb', 'nx', 'ny'),
                         x0, ('nb', 'nx', 'ny'),
                         beam, bout,
                         lastsize, None,
                         nthreads, None,
                         sigmainv, None,
                         cgopts, None,
                         align_arrays=False,
                         dtype=b.dtype)
    if compute:
        return model.compute()
    else:
        return model


def pcg_dist(A, maxit, minit, tol, sigmainv):
    '''
    kwargs - tol, maxit, minit, nthreads, psf_padding, unpad_x, unpad_y
             sigmainv, lastsize, hessian
    '''

    if hasattr(A, 'residual'):
        b = A.residual/A.wsum
    else:
        b = A.dirty/A.wsum
    x = np.zeros_like(b)

    def M(x):
        return x / sigmainv

    r = A(x) - b
    y = M(r)
    p = -y
    rnorm = np.vdot(r, y)
    if np.isnan(rnorm) or rnorm == 0.0:
        eps0 = 1.0
    else:
        eps0 = rnorm
    k = 0
    eps = 1.0
    stall_count = 0
    while (eps > tol or k < minit) and k < maxit and stall_count < 5:
        xp = x.copy()
        rp = r.copy()
        epsp = eps
        Ap = A(p)
        rnorm = np.vdot(r, y)
        alpha = rnorm / np.vdot(p, Ap)
        x = xp + alpha * p
        r = rp + alpha * Ap
        y = M(r)
        rnorm_next = np.vdot(r, y)
        while rnorm_next > rnorm:  # TODO - better line search
            alpha *= 0.75
            x = xp + alpha * p
            r = rp + alpha * Ap
            y = M(r)
            rnorm_next = np.vdot(r, y)

        beta = rnorm_next / rnorm
        p = beta * p - y
        # if p is zero we should stop
        if not np.any(p):
            break
        rnorm = rnorm_next
        k += 1
        eps = rnorm / eps0

        if np.abs(eps - epsp) < 1e-3*tol:
            stall_count += 1

    print(f'Band={A.bandid}, iters{k}, eps={eps}', file=log)
    return x





'''
Shameless pillaged from
https://github.com/tart-telescope/tart2ms/blob/master/tart2ms/fixvis.py
'''

from pyrap.tables import table as tbl
import numpy as np
from pyrap.measures import measures
from pyrap.quanta import quantity
from astropy.time import Time
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates import solar_system_ephemeris, EarthLocation, AltAz
from astropy.coordinates import get_body_barycentric, get_body, get_moon

def synthesize_uvw(station_ECEF, time, a1, a2,
                   phase_ref,
                   stopctr_units=["rad", "rad"], stopctr_epoch="j2000",
                   time_TZ="UTC", time_unit="s",
                   posframe="ITRF", posunits=["m", "m", "m"]):
    """
    Synthesizes new UVW coordinates based on time according to
    NRAO CASA convention (same as in fixvis)
    station_ECEF: ITRF station coordinates read from MS::ANTENNA
    time: time column, preferably time centroid
    a1: ANTENNA_1 index
    a2: ANTENNA_2 index
    phase_ref: phase reference centre in radians
    returns dictionary of dense uvw coordinates and indices:
        {
         "UVW": shape (nbl * ntime, 3),
         "TIME_CENTROID": shape (nbl * ntime,),
         "ANTENNA_1": shape (nbl * ntime,),
         "ANTENNA_2": shape (nbl * ntime,)
        }
    Note: input and output antenna indexes may not have the same
          order or be flipped in 1 to 2 index
    Note: This operation CANNOT be applied blockwise due
          to a casacore.measures threadsafety issue
    """
    assert time.size == a1.size
    assert a1.size == a2.size

    ants = np.concatenate((a1, a2))
    unique_ants = np.arange(np.max(ants) + 1)
    unique_time = np.unique(time)
    na = unique_ants.size
    nbl = na * (na - 1) // 2 + na
    ntime = unique_time.size
    # keep a full uvw array for all antennae - including those
    # dropped by previous calibration and CASA splitting
    padded_uvw = np.zeros((ntime * nbl, 3), dtype=np.float64)
    antindices = np.stack(np.triu_indices(na, 0),
                          axis=1)
    padded_time = unique_time.repeat(nbl)
    padded_a1 = np.tile(antindices[:, 0], (1, ntime)).ravel()
    padded_a2 = np.tile(antindices[:, 1], (1, ntime)).ravel()

    dm = measures()
    epoch = dm.epoch(time_TZ, quantity(time[0], time_unit))
    refdir = dm.direction(stopctr_epoch,
                          quantity(phase_ref[0, 0], stopctr_units[0]),
                          quantity(phase_ref[0, 1], stopctr_units[1]))
    obs = dm.position(posframe,
                      quantity(station_ECEF[0, 0], posunits[0]),
                      quantity(station_ECEF[0, 1], posunits[1]),
                      quantity(station_ECEF[0, 2], posunits[2]))

    #setup local horizon coordinate frame with antenna 0 as reference position
    dm.do_frame(obs)
    dm.do_frame(refdir)
    dm.do_frame(epoch)
    for ti, t in enumerate(unique_time):
        epoch = dm.epoch("UT1", quantity(t, "s"))
        dm.do_frame(epoch)

        station_uv = np.zeros_like(station_ECEF)
        for iapos, apos in enumerate(station_ECEF):
            compuvw = dm.to_uvw(dm.baseline(posframe,
                                            quantity([apos[0], station_ECEF[0, 0]], posunits[0]),
                                            quantity([apos[1], station_ECEF[0, 1]], posunits[1]),
                                            quantity([apos[2], station_ECEF[0, 2]], posunits[2])))
            station_uv[iapos] = compuvw["xyz"].get_value()[0:3]
        for bl in range(nbl):
            blants = antindices[bl]
            bla1 = blants[0]
            bla2 = blants[1]
            # same as in CASA convention (Convention for UVW calculations in CASA, Rau 2013)
            padded_uvw[ti*nbl + bl, :] = station_uv[bla1] - station_uv[bla2]

    # # hack to remove auto-correlations
    # if a1.size != padded_a1.size:
    #     I = np.where(padded_a1 != padded_a2)
    #     padded_uvw = padded_uvw[I]
    #     padded_time = padded_time[I]
    #     padded_a1 = padded_a1[I]
    #     padded_a2 = padded_a2[I]

    return dict(zip(["UVW", "TIME_CENTROID", "ANTENNA1", "ANTENNA2"],
                    [padded_uvw, padded_time, padded_a1, padded_a2]))


# Pirated from
# https://github.com/ratt-ru/solarkat/blob/main/solarkat-pipeline/find_sun_stimela.py
def get_coordinates(obs_time,
                    obs_lat=-30.71323598930457,
                    obs_lon=21.443001467965008,
                    target='Sun'):
    '''
    Give location of object given telescope location (defaults to MeerKAT)
    and time of observation.

    Inputs:

    obs_time - should be weighted mean of TIME from MS
    obs_lat  - telescope lattitude in degrees
    obs_lon  - telescope longitude in degrees
    '''
    def format_coords(ra0,dec0):
        c = SkyCoord(ra0*u.deg,dec0*u.deg,frame='fk5')
        hms = str(c.ra.to_string(u.hour))
        dms = str(c.dec)
        return hms,dms


    loc = EarthLocation.from_geodetic(obs_lat,obs_lon) #,obs_height,ellipsoid)
    t = Time(obs_time/86400.0,format='mjd')  # where is this factor from?
    with solar_system_ephemeris.set('builtin'):
        sun = get_body(target, t, loc)
        sun_ra = sun.ra.value
        sun_dec = sun.dec.value
    sun_hms=format_coords(sun_ra,sun_dec)
    # print(sun_hms[0],sun_hms[1])
    return np.deg2rad(sun_ra), np.deg2rad(sun_dec)



# import packratt
# import pytest
# from pathlib import Path
# from xarray import Dataset
# from collections import namedtuple
# import dask
# import dask.array as da
# from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
# pmp = pytest.mark.parametrize

# @pmp('beam_model', (None, 'kbl',))
# @pmp('do_gains', (False, True))
# def test_fwdbwd(beam_model, do_gains, tmp_path_factory):
#     '''
#     Here we test that the forward backward algorithm correctly infers
#     the fluxes of point sources with known locations in the presence
#     of the wterm, DI gain corruptions and a static known primary beam.
#     TODO - add per scan PB variations
#     '''
#     test_dir = tmp_path_factory.mktemp("test_pfb")
#     # test_dir = Path('/home/landman/data/')
#     packratt.get('/test/ms/2021-06-24/elwood/test_ascii_1h60.0s.MS.tar', str(test_dir))

#     import numpy as np
#     np.random.seed(420)
#     from numpy.testing import assert_allclose
#     from pyrap.tables import table
#     from africanus.constants import c as lightspeed
#     from ducc0.fft import good_size
#     from katbeam import JimBeam
#     from ducc0.wgridder import dirty2ms
#     from africanus.gps.utils import abs_diff
#     from pfb.utils.misc import kron_matvec
#     from pfb.utils.misc import chunkify_rows
#     from africanus.calibration.utils import corrupt_vis
#     from pfb.utils.fits import save_fits, set_wcs

#     ms = table(str(test_dir / 'test_ascii_1h60.0s.MS'), readonly=False)
#     spw = table(str(test_dir / 'test_ascii_1h60.0s.MS::SPECTRAL_WINDOW'))
#     fields = xds_from_table(str(test_dir / 'test_ascii_1h60.0s.MS::FIELD'))
#     radec = fields[0].PHASE_DIR.data.squeeze()

#     utime = np.unique(ms.getcol('TIME'))

#     freq = spw.getcol('CHAN_FREQ').squeeze()
#     freq0 = np.mean(freq)

#     ntime = utime.size
#     nchan = freq.size
#     nant = np.maximum(ms.getcol('ANTENNA1').max(), ms.getcol('ANTENNA2').max()) + 1

#     ncorr = ms.getcol('FLAG').shape[-1]

#     uvw = ms.getcol('UVW')
#     nrow = uvw.shape[0]
#     u_max = abs(uvw[:, 0]).max()
#     v_max = abs(uvw[:, 1]).max()
#     uv_max = np.maximum(u_max, v_max)

#     # image size
#     cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

#     srf = 2.0
#     cell_rad = cell_N / srf
#     cell_deg = cell_rad * 180 / np.pi
#     cell_size = cell_deg * 3600
#     print("Cell size set to %5.5e arcseconds" % cell_size)

#     # the test will fail in intrinsic if sources fall near beam sidelobes
#     fov = 1.0
#     npix = good_size(int(fov / cell_deg))
#     while npix % 2:
#         npix += 1
#         npix = good_size(npix)

#     nx = npix
#     ny = npix

#     print("Image size set to (%i, %i, %i)" % (nchan, nx, ny))

#     # model
#     model = np.zeros((nchan, nx, ny), dtype=np.float64)
#     nsource = 10
#     Ix = np.random.randint(0, npix, nsource)
#     Iy = np.random.randint(0, npix, nsource)
#     alpha = -0.7 + 0.1 * np.random.randn(nsource)
#     I0 = 1.0 + np.abs(np.random.randn(nsource))
#     for i in range(nsource):
#         model[:, Ix[i], Iy[i]] = I0[i] * (freq/freq0) ** alpha[i]

#     # TODO - interpolate beam
#     if beam_model is None:
#         beam = np.ones((nchan, nx, ny), dtype=float)
#     elif beam_model == 'kbl':
#         beamo = JimBeam('MKAT-AA-L-JIM-2020').I
#         l = (-(nx//2) + np.arange(nx)) * cell_deg
#         m = (-(ny//2) + np.arange(ny)) * cell_deg
#         ll, mm = np.meshgrid(l, m, indexing='ij')
#         beam = np.zeros((nchan, nx, ny), dtype=float)
#         for c in range(nchan):
#             beam[c] = beamo(ll, mm, freq[c]/1e6)

#     # model vis
#     epsilon = 1e-7  # tests take too long if smaller
#     model_vis = np.zeros((nrow, nchan, ncorr), dtype=np.complex128)
#     for c in range(nchan):
#         model_vis[:, c:c+1, 0] = dirty2ms(uvw, freq[c:c+1], beam[c]*model[c],
#                                     pixsize_x=cell_rad, pixsize_y=cell_rad,
#                                     epsilon=epsilon, do_wstacking=True, nthreads=8)
#         model_vis[:, c, -1] = model_vis[:, c, 0]

#     desc = ms.getcoldesc('DATA')
#     desc['name'] = 'DATA2'
#     desc['valueType'] = 'dcomplex'
#     desc['comment'] = desc['comment'].replace(" ", "_")
#     dminfo = ms.getdminfo('DATA')
#     dminfo["NAME"] =  "{}-{}".format(dminfo["NAME"], 'DATA2')
#     ms.addcols(desc, dminfo)

#     if do_gains:
#         t = (utime-utime.min())/(utime.max() - utime.min())
#         nu = 2*(freq/freq0 - 1.0)

#         tt = abs_diff(t, t)
#         lt = 0.25
#         Kt = 0.1 * np.exp(-tt**2/(2*lt**2))
#         Lt = np.linalg.cholesky(Kt + 1e-10*np.eye(ntime))
#         vv = abs_diff(nu, nu)
#         lv = 0.1
#         Kv = 0.1 * np.exp(-vv**2/(2*lv**2))
#         Lv = np.linalg.cholesky(Kv + 1e-10*np.eye(nchan))
#         L = (Lt, Lv)

#         jones = np.zeros((ntime, nchan, nant, 1, ncorr), dtype=np.complex128)
#         for p in range(nant):
#             for c in [0, -1]:  # for now only diagonal
#                 xi_amp = np.random.randn(ntime, nchan)
#                 amp = np.exp(-nu[None, :]**2 + kron_matvec(L, xi_amp))
#                 xi_phase = np.random.randn(ntime, nchan)
#                 phase = kron_matvec(L, xi_phase)
#                 jones[:, :, p, 0, c] = amp * np.exp(1.0j * phase)


#         # corrupted vis
#         model_vis = model_vis.reshape(nrow, nchan, 1, 2, 2)
#         time = ms.getcol('TIME')
#         row_chunks, tbin_idx, tbin_counts = chunkify_rows(time, ntime)
#         ant1 = ms.getcol('ANTENNA1')
#         ant2 = ms.getcol('ANTENNA2')

#         vis = corrupt_vis(tbin_idx, tbin_counts, ant1, ant2,
#                           np.swapaxes(jones, 1, 2), model_vis).reshape(nrow, nchan, ncorr)
#         ms.putcol('DATA2', vis)

#         # cast gain to QuartiCal format
#         g = da.from_array(jones)
#         gflags = da.zeros((ntime, nchan, nant, 1))
#         data_vars = {
#             'gains':(('gain_time', 'gain_freq', 'antenna', 'direction', 'correlation'), g),
#             'gain_flags':(('gain_time', 'gain_freq', 'antenna', 'direction'), gflags)
#         }
#         gain_spec_tup = namedtuple('gains_spec_tup', 'tchunk fchunk achunk dchunk cchunk')
#         attrs = {
#             'NAME': 'NET',
#             'TYPE': 'complex',
#             'FIELD_NAME': '00',
#             'SCAN_NUMBER': int(1),
#             'FIELD_ID': int(0),
#             'DATA_DESC_ID': int(0),
#             'GAIN_SPEC': gain_spec_tup(tchunk=(int(ntime),),
#                                        fchunk=(int(nchan),),
#                                        achunk=(int(nant),),
#                                        dchunk=(int(1),),
#                                        cchunk=(int(ncorr),)),
#             'GAIN_AXES': ('gain_time', 'gain_freq', 'antenna', 'direction', 'correlation')
#         }
#         coords = {
#             'gain_freq': (('gain_freq',), freq),
#             'gain_time': (('gain_time',), utime)

#         }
#         net_xds_list = Dataset(data_vars, coords=coords, attrs=attrs)
#         gain_path = str(test_dir / Path("gains.qc"))
#         out_path = f'{gain_path}::NET'
#         dask.compute(xds_to_zarr(net_xds_list, out_path))

#     else:
#         ms.putcol('DATA2', model_vis)
#         gain_path = None


#     postfix = ""
#     # set defaults from schema
#     from pfb.parser.schemas import schema
#     init_args = {}
#     for key in schema.init["inputs"].keys():
#         init_args[key] = schema.init["inputs"][key]["default"]
#     # overwrite defaults
#     outname = str(test_dir / 'test')
#     init_args["ms"] = str(test_dir / 'test_ascii_1h60.0s.MS')
#     init_args["output_filename"] = outname
#     init_args["nband"] = nchan
#     init_args["data_column"] = "DATA2"
#     init_args["flag_column"] = 'FLAG'
#     init_args["gain_table"] = gain_path
#     init_args["beam_model"] = beam_model
#     init_args["max_field_of_view"] = fov
#     from pfb.workers.init import _init
#     _init(**init_args)

#     # grid data to produce dirty image
#     grid_args = {}
#     for key in schema.grid["inputs"].keys():
#         grid_args[key] = schema.grid["inputs"][key]["default"]
#     # overwrite defaults
#     grid_args["output_filename"] = outname
#     grid_args["postfix"] = postfix
#     grid_args["nband"] = nchan
#     grid_args["field_of_view"] = fov
#     grid_args["fits_mfs"] = False
#     grid_args["psf"] = True
#     grid_args["nthreads"] = 8  # has to be set when calling _grid
#     grid_args["nvthreads"] = 8
#     grid_args["overwrite"] = True
#     grid_args["robustness"] = None
#     grid_args["wstack"] = True
#     grid_args["residual"] = False
#     from pfb.workers.grid import _grid
#     _grid(**grid_args)

#     # place mask in mds
#     mask = np.any(model, axis=0)
#     maskname = f'{outname}_I.mask.fits'
#     hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freq[0])
#     save_fits(maskname, mask, hdr)

#     # run fwdbwd to get model
#     fwdbwd_args = {}
#     for key in schema.fwdbwd["inputs"].keys():
#         fwdbwd_args[key] = schema.fwdbwd["inputs"][key]["default"]
#     fwdbwd_args["output_filename"] = outname
#     fwdbwd_args["postfix"] = postfix
#     fwdbwd_args["nband"] = nchan
#     fwdbwd_args["mask"] = maskname
#     fwdbwd_args["nthreads"] = 8  # has to be set when calling _fwdbwd
#     fwdbwd_args["nvthreads"] = 8
#     fwdbwd_args["pm_tol"] = 0.1*epsilon
#     fwdbwd_args["pm_maxit"] = 250
#     fwdbwd_args["cg_tol"] = 0.1*epsilon
#     fwdbwd_args["cg_maxit"] = 50
#     fwdbwd_args["cg_minit"] = 1
#     fwdbwd_args["pd_tol"] = 0.1*epsilon
#     fwdbwd_args["pd_maxit"] = 100
#     fwdbwd_args["wstack"] = True
#     fwdbwd_args["sigma21"] = 0
#     fwdbwd_args["alpha"] = 1
#     fwdbwd_args["bases"] = 'self'
#     fwdbwd_args["sigmainv"] = 0.0
#     fwdbwd_args["tol"] = 0.1*epsilon
#     fwdbwd_args["niter"] = 10
#     fwdbwd_args["fits_mfs"] = False

#     from pfb.workers.fwdbwd import _fwdbwd
#     _fwdbwd(**fwdbwd_args)

#     # get inferred model
#     basename = f'{outname}_I'
#     dds_name = f'{basename}{postfix}.dds.zarr'
#     dds = xds_from_zarr(dds_name, chunks={'x':-1, 'y': -1})
#     model_inferred = np.zeros((nchan, nx, ny))
#     for ds in dds:
#         b = ds.bandid
#         model_inferred[b] = ds.MODEL.values

#     # we do not expect a perfect match after a handful of iterations
#     # hence larger tolerance of 5e-6
#     for i in range(nsource):
#         assert_allclose(1.0 + model_inferred[:, Ix[i], Iy[i]] -
#                         model[:, Ix[i], Iy[i]], 1.0,
#                         atol=50*epsilon, rtol=50*epsilon)

#  # beam_model, do_gains
# # test_fwdbwd('kbl', True)



import numpy as np
import numexpr as ne
from functools import partial
import numba
import dask.array as da
from pfb.operators.psf import psf_convolve_cube
from ducc0.misc import make_noncritical
import pyscilog
log = pyscilog.get_logger('CLARK')

@numba.jit(parallel=True, nopython=True, nogil=True, cache=True, inline='always')
def subtract(A, psf, Ip, Iq, xhat, nxo2, nyo2):
    '''
    Subtract psf centered at location of xhat
    '''
    # loop over active indices
    nband = xhat.size
    for b in numba.prange(nband):
    # for b in range(nband):
        for i in range(Ip.size):
            pp = nxo2 - Ip[i]
            qq = nyo2 - Iq[i]
            A[b, i] -= xhat[b] * psf[b, pp, qq]
            # print(b, pp, qq, psf[b, pp, qq])
    return A


@numba.jit(parallel=True, nopython=True, nogil=True, cache=True)
def subminor(A, psf, Ip, Iq, model, wsums, gamma=0.05, th=0.0, maxit=10000):
    """
    Run subminor loop in active set

    A       - active set (all pixels above subminor threshold)
    psf     - psf image
    Ip      - x indices of active set
    Iq      - y indices of active set
    model   - current model image
    gamma   - loop gain
    th      - threshold to clean down to
    maxit   - maximum number of iterations

    Pixels in A map to pixels in the image via

    p = Ip[pq]
    q = Iq[pq]

    where pq is the location of the maximum in A.
    """
    nband, nx_psf, ny_psf = psf.shape
    nxo2 = nx_psf//2
    nyo2 = ny_psf//2
    Asearch = np.sum(A, axis=0)**2
    pq = Asearch.argmax()
    p = Ip[pq]
    q = Iq[pq]
    Amax = np.sqrt(Asearch[pq])
    fsel = wsums > 0
    k = 0
    while Amax > th and k < maxit:
        xhat = A[:, pq]
        model[fsel, p, q] += gamma * xhat[fsel]/wsums[fsel]
        Idelp = p - Ip
        Idelq = q - Iq
        # find where PSF overlaps with image
        mask = (np.abs(Idelp) <= nxo2) & (np.abs(Idelq) <= nyo2)
        # Ipp, Iqq = psf[:, nxo2 - Ip[mask], nyo2 - Iq[mask]]
        A = subtract(A[:, mask], psf,
                     Idelp[mask], Idelq[mask],
                     xhat, nxo2, nyo2)

        Asearch = np.sum(A, axis=0)**2
        pq = Asearch.argmax()
        p = Ip[pq]
        q = Iq[pq]
        Amax = np.sqrt(Asearch[pq])
        k += 1
    return model

def clark(ID,
          PSF,
          PSFHAT,
          wsums,
          threshold=0,
          gamma=0.05,
          pf=0.05,
          maxit=50,
          subpf=0.5,
          submaxit=1000,
          report_freq=1,
          verbosity=1,
          psfopts=None,
          sigmathreshold=2,
          nthreads=1):
    nband, nx, ny = ID.shape
    _, nx_psf, ny_psf = PSF.shape
    # we assume that the dirty image and PSF have been normalised by wsum
    # and that we get units of Jy/beam when we take the sum over the frequency
    # axis i.e. the MFS image is in units of Jy/beam
    wsum = wsums.sum()
    assert np.allclose(wsum, 1)
    # ID /= wsum
    # PSF /= wsum
    # PSFHAT /= wsum
    # wsums = wsums/wsum
    model = np.zeros((nband, nx, ny), dtype=ID.dtype)
    model = make_noncritical(model)
    IR = ID.copy()
    # pre-allocate arrays for doing FFT's
    xout = np.empty(ID.shape, dtype=ID.dtype, order='C')
    xout = make_noncritical(xout)
    xpad = np.empty(PSF.shape, dtype=ID.dtype, order='C')
    xpad = make_noncritical(xpad)
    xhat = np.empty(PSFHAT.shape, dtype=PSFHAT.dtype)
    xhat = make_noncritical(xhat)
    # square avoids abs of full array
    IRsearch = np.sum(IR, axis=0)**2
    pq = IRsearch.argmax()
    p = pq//ny
    q = pq - p*ny
    IRmax = np.sqrt(IRsearch[p, q])
    tol = np.maximum(pf * IRmax, threshold)
    k = 0
    stall_count = 0
    while IRmax > tol and k < maxit and stall_count < 5:
        # identify active set
        subth = subpf * IRmax
        Ip, Iq = np.where(IRsearch > subth**2)
        # run substep in active set
        model = subminor(IR[:, Ip, Iq], PSF, Ip, Iq, model, wsums,
                         gamma=gamma,
                         th=subth,
                         maxit=submaxit)
        # subtract from full image (as in major cycle)
        psf_convolve_cube(
                        xpad,
                        xhat,
                        xout,
                        PSFHAT,
                        ny_psf,
                        model,
                        nthreads=nthreads)
        IR = ID - xout
        IRsearch = np.sum(IR, axis=0)**2
        pq = IRsearch.argmax()
        p = pq//ny
        q = pq - p*ny
        IRmaxp = IRmax
        IRmax = np.sqrt(IRsearch[p, q])
        k += 1

        if np.abs(IRmaxp - IRmax) / np.abs(IRmaxp) < 1e-3:
            stall_count += stall_count

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} max resid = {IRmax}",
                  file=log)

    IRmfs = np.sum(IR, axis=0)
    rms = np.std(IRmfs[~np.any(model, axis=0)])

    if k >= maxit:
        if verbosity:
            print(f"Max iters reached. "
                  f"Max resid = {IRmax:.3e}, rms = {rms:.3e}", file=log)
        return model, 1
    elif stall_count >= 5:
        if verbosity:
            print(f"Stalled. "
                  f"Max resid = {IRmax:.3e}, rms = {rms:.3e}", file=log)
        return model, 1
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations. "
                  f"Max resid = {IRmax:.3e}, rms = {rms:.3e}", file=log)
        return model, 0



from collections import namedtuple
import warnings

import numba
import numba.core.types as nbtypes
from numba.core.cgutils import is_nonelike
from numba.cpython.unsafe.tuple import tuple_setitem
from numba.extending import register_jitable, overload
from numba.typed import Dict, List
import numpy as np

from pfb.wavelets.coefficients import coefficients
from pfb.wavelets.common import NUMBA_SEQUENCE_TYPES
from pfb.wavelets.convolution import downsampling_convolution
from pfb.wavelets.convolution import upsampling_convolution_valid_sf
from pfb.wavelets.modes import Modes, promote_mode
from pfb.wavelets.intrinsics import slice_axis, force_type_contiguity, not_optional


BaseWavelet = namedtuple("BaseWavelet", (
    "support_width",
    "symmetry",
    "orthogonal",
    "biorthogonal",
    "compact_support",
    "family_name",
    "short_name"))


DiscreteWavelet = namedtuple("DiscreteWavelet",
                             BaseWavelet._fields + (
                                 "dec_hi",
                                 "dec_lo",
                                 "rec_hi",
                                 "rec_lo",
                                 "vanishing_moments_psi",
                                 "vanishing_moments_phi"))


@register_jitable
def str_to_int(s):
    final_index = len(s) - 1
    value = 0

    for i, v in enumerate(s):
        digit = (ord(v) - 48)

        if digit < 0 or digit > 9:
            raise ValueError("Invalid integer string")

        value += digit * (10 ** (final_index - i))

    return value


@register_jitable
def dwt_max_level(data_length, filter_length):
    if filter_length < 2:
        raise ValueError("Invalid wavelet filter length")

    if filter_length <= 1 or data_length < (filter_length - 1):
        return 0

    return int(np.floor(np.log2(data_length / (filter_length - 1))))


@register_jitable
def dwt_buffer_length(input_length, filter_length, mode):
    if input_length < 1 or filter_length < 1:
        return 0

    if mode is Modes.periodisation:
        return (input_length // 2) + (1 if input_length % 2 else 0)
    else:
        return (input_length + filter_length - 1) // 2


@register_jitable
def idwt_buffer_length(coeffs_length, filter_length, mode):
    if mode is Modes.periodisation:
        return 2 * coeffs_length
    else:
        return 2 * coeffs_length - filter_length + 2


@register_jitable
def dwt_coeff_length(data_length, filter_length, mode):
    if data_length < 1:
        raise ValueError("data_length < 1")

    if filter_length < 1:
        raise ValueError("filter_length < 1")

    return dwt_buffer_length(data_length, filter_length, mode)


@numba.generated_jit(nopython=True, nogil=True, cache=True)
def discrete_wavelet(wavelet):
    if not isinstance(wavelet, nbtypes.types.UnicodeType):
        raise TypeError("wavelet must be a string")

    def impl(wavelet):
        if wavelet.startswith("db"):
            offset = 2
            order = str_to_int(wavelet[offset:])

            coeffs = coefficients("db", order - 1)
            index = np.arange(len(coeffs))
            rec_lo = coeffs
            dec_lo = np.flipud(rec_lo)
            rec_hi = np.where(index % 2, -1, 1) * dec_lo
            dec_hi = np.where(index[::-1] % 2, -1, 1) * rec_lo

            return DiscreteWavelet(support_width=2 * order - 1,
                                   symmetry="asymmetric",
                                   orthogonal=1,
                                   biorthogonal=1,
                                   compact_support=1,
                                   family_name="Daubechies",
                                   short_name="db",
                                   rec_lo=rec_lo,
                                   dec_lo=dec_lo,
                                   rec_hi=rec_hi,
                                   dec_hi=dec_hi,
                                   vanishing_moments_psi=order,
                                   vanishing_moments_phi=0)

        else:
            raise ValueError("Unknown wavelet")

    return impl


@numba.generated_jit(nopython=True, nogil=True, cache=True)
def promote_wavelets(wavelets, naxis):
    if not isinstance(naxis, nbtypes.Integer):
        raise TypeError("naxis must be an integer")

    if isinstance(wavelets, nbtypes.misc.UnicodeType):
        def impl(wavelets, naxis):
            return List([wavelets] * naxis)

    elif (isinstance(wavelets, NUMBA_SEQUENCE_TYPES) and
            isinstance(wavelets.dtype, nbtypes.misc.UnicodeType)):

        def impl(wavelets, naxis):
            if len(wavelets) != naxis:
                raise ValueError("len(wavelets) != len(axis)")

            return List(wavelets)
    else:
        raise TypeError("wavelet must be a string, "
                        "a list of strings "
                        "or a tuple of strings. "
                        "Got %s." % wavelets)

    return impl


@numba.generated_jit(nopython=True, nogil=True, cache=True)
def promote_axis(axis, ndim):
    if not isinstance(ndim, nbtypes.Integer):
        raise TypeError("ndim must be an integer")

    if isinstance(axis, nbtypes.Integer):
        def impl(axis, ndim):
            axis = axis + ndim if axis < 0 else axis
            return List([axis])

    elif (isinstance(axis, NUMBA_SEQUENCE_TYPES) and
            isinstance(axis.dtype, nbtypes.Integer)):
        def impl(axis, ndim):
            if len(axis) > ndim:
                raise ValueError("len(axis) > data.ndim")

            for a in axis:
                if a >= ndim:
                    raise ValueError("axis[i] >= data.ndim")

            return List([a + ndim if a < 0 else a for a in axis])

    else:
        raise TypeError("axis must be an integer, "
                        "a list of integers "
                        "or a tuple of integers. "
                        "Got %s." % axis)

    return impl


@numba.generated_jit(nopython=True, nogil=True, fastmath=True, cache=True)
def dwt_axis(data, wavelet, mode, axis):
    def impl(data, wavelet, mode, axis):
        coeff_len = dwt_coeff_length(
            data.shape[axis], len(
                wavelet.dec_hi), mode)
        out_shape = tuple_setitem(data.shape, axis, coeff_len)

        if axis < 0 or axis >= data.ndim:
            raise ValueError("0 <= axis < data.ndim failed")

        ca = np.empty(out_shape, dtype=data.dtype)
        cd = np.empty(out_shape, dtype=data.dtype)

        # Iterate over all points except along the slicing axis
        for idx in np.ndindex(*tuple_setitem(data.shape, axis, 1)):
            initial_in_row = slice_axis(data, idx, axis, None)
            initial_ca_row = slice_axis(ca, idx, axis, None)
            initial_cd_row = slice_axis(cd, idx, axis, None)

            # The numba array type returned by slice_axis assumes
            # non-contiguity in the general case.
            # However, the slice may actually be contiguous in layout
            # If so, cast the array type to obtain type contiguity
            # else, copy the slice to obtain contiguity in
            # both type and layout
            if initial_in_row.flags.c_contiguous:
                in_row = force_type_contiguity(initial_in_row)
            else:
                in_row = initial_in_row.copy()

            if initial_ca_row.flags.c_contiguous:
                ca_row = force_type_contiguity(initial_ca_row)
            else:
                ca_row = initial_ca_row.copy()

            if initial_cd_row.flags.c_contiguous:
                cd_row = force_type_contiguity(initial_cd_row)
            else:
                cd_row = initial_cd_row.copy()

            # Compute the approximation and detail coefficients
            downsampling_convolution(in_row, ca_row, wavelet.dec_lo, mode, 2)
            downsampling_convolution(in_row, cd_row, wavelet.dec_hi, mode, 2)

            # If necessary, copy back into the output
            if not initial_ca_row.flags.c_contiguous:
                initial_ca_row[:] = ca_row[:]

            if not initial_cd_row.flags.c_contiguous:
                initial_cd_row[:] = cd_row[:]

        return ca, cd

    return impl


@numba.generated_jit(nopython=True, nogil=True, fastmath=True, cache=True)
def idwt_axis(approx_coeffs, detail_coeffs,
              wavelet, mode, axis):

    have_approx = not is_nonelike(approx_coeffs)
    have_detail = not is_nonelike(detail_coeffs)

    if not have_approx and not have_detail:
        raise ValueError("Either approximation or detail "
                         "coefficients must be present")

    dtypes = [approx_coeffs.dtype if have_approx else None,
              detail_coeffs.dtype if have_detail else None]

    out_dtype = np.result_type(*(np.dtype(dt.name) for dt in dtypes if dt))

    if have_approx and have_detail:
        if approx_coeffs.ndim != detail_coeffs.ndim:
            raise ValueError("approx_coeffs.ndim != detail_coeffs.ndim")

    def impl(approx_coeffs, detail_coeffs,
             wavelet, mode, axis):

        if have_approx and have_detail:
            coeff_shape = approx_coeffs.shape
            it = enumerate(zip(approx_coeffs.shape, detail_coeffs.shape))

            # NOTE(sjperkins)
            # Clip the coefficient dimensions to the smallest dimensions
            # pywt clips in waverecn and fails in idwt and idwt_axis
            # on heterogenous coefficient shapes.
            # The actual clipping is performed in slice_axis
            for i, (asize, dsize) in it:
                size = asize if asize < dsize else dsize
                coeff_shape = tuple_setitem(coeff_shape, i, size)

        elif have_approx:
            coeff_shape = approx_coeffs.shape
        elif have_detail:
            coeff_shape = detail_coeffs.shape
        else:
            raise ValueError("Either approximation or detail must be present")

        if not (0 <= axis < len(coeff_shape)):
            raise ValueError(("0 <= axis < coeff.ndim does not hold"))

        idwt_len = idwt_buffer_length(coeff_shape[axis],
                                      wavelet.rec_lo.shape[0],
                                      mode)
        out_shape = tuple_setitem(coeff_shape, axis, idwt_len)
        output = np.empty(out_shape, dtype=out_dtype)

        # Iterate over all points except along the slicing axis
        for idx in np.ndindex(*tuple_setitem(output.shape, axis, 1)):
            initial_out_row = slice_axis(output, idx, axis, None)

            # Zero if we have a contiguous slice, else allocate
            if initial_out_row.flags.c_contiguous:
                out_row = force_type_contiguity(initial_out_row)
                out_row[:] = 0
            else:
                out_row = np.zeros_like(initial_out_row)

            # Apply approximation coefficients if they exist
            if approx_coeffs is not None:
                initial_ca_row = slice_axis(approx_coeffs, idx,
                                            axis, coeff_shape[axis])

                if initial_ca_row.flags.c_contiguous:
                    ca_row = force_type_contiguity(initial_ca_row)
                else:
                    ca_row = initial_ca_row.copy()

                upsampling_convolution_valid_sf(ca_row, wavelet.rec_lo,
                                                out_row, mode)

            # Apply detail coefficients if they exist
            if detail_coeffs is not None:
                initial_cd_row = slice_axis(detail_coeffs, idx,
                                            axis, coeff_shape[axis])

                if initial_cd_row.flags.c_contiguous:
                    cd_row = force_type_contiguity(initial_cd_row)
                else:
                    cd_row = initial_cd_row.copy()

                upsampling_convolution_valid_sf(cd_row, wavelet.rec_hi,
                                                out_row, mode)

            # Copy back output row if the output space was non-contiguous
            if not initial_out_row.flags.c_contiguous:
                initial_out_row[:] = out_row

        return output

    return impl


@numba.generated_jit(nopython=True, nogil=True, fastmath=True, cache=True)
def dwt(data, wavelet, mode="zero", axis=None):

    if isinstance(data, nbtypes.misc.Optional):
        if not isinstance(data.type, nbtypes.npytypes.Array):
            raise TypeError(f"data must be ndarray. Got {data.type}")
    elif not isinstance(data, nbtypes.npytypes.Array):
        raise TypeError(f"data must be an ndarray. Got {data}")

    have_axis = not is_nonelike(axis)

    def impl(data, wavelet, mode="zero", axis=None):
        if not have_axis:
            axis = List(range(data.ndim))

        paxis = promote_axis(axis, data.ndim)
        naxis = len(paxis)
        pmode = promote_mode(mode, naxis)
        pwavelets = [discrete_wavelet(w) for w
                     in promote_wavelets(wavelet, naxis)]

        coeffs = List([("", data)])

        for a, (ax, m, wv) in enumerate(zip(paxis, pmode, pwavelets)):
            new_coeffs = List()

            for subband, x in coeffs:
                ca, cd = dwt_axis(x, wv, m, ax)
                new_coeffs.append((subband + "a", ca))
                new_coeffs.append((subband + "d", cd))

            coeffs = new_coeffs

        dict_coeffs = Dict()

        for name, coeff in coeffs:
            dict_coeffs[name] = coeff

        return dict_coeffs

    return impl


@register_jitable
def coeff_product(args, repeat=1):
    # Adapted from
    # https://docs.python.org/3/library/itertools.html#itertools.product
    pools = List()

    for i in range(repeat):
        pools.append(args)

    result = List()

    for i in range(len(args) - 1):
        result.append('')

    for pool in pools:
        result = List([x + y for x in result for y in pool])

    return result


@numba.generated_jit(nopython=True, nogil=True, fastmath=True, cache=True)
def idwt(coeffs, wavelet, mode='zero', axis=None):

    have_axis = not is_nonelike(axis)

    def impl(coeffs, wavelet, mode='zero', axis=None):
        ndim_transform = max([len(key) for key in coeffs.keys()])
        coeff_shapes = [v.shape for v in coeffs.values()]

        if not have_axis:
            axis = List(range(ndim_transform))
            ndim = ndim_transform
        else:
            ndim = len(coeff_shapes[0])

        paxis = promote_axis(axis, ndim)
        naxis = len(paxis)
        pmode = promote_mode(mode, naxis)
        pwavelets = List([discrete_wavelet(w) for w
                          in promote_wavelets(wavelet, naxis)])

        it = list(enumerate(zip(paxis, pwavelets, pmode)))

        for key_length, (ax, wv, m) in it[::-1]:
            new_coeffs = {}
            new_keys = coeff_product('ad', key_length)

            for key in new_keys:
                L = coeffs[key + 'a']
                H = coeffs[key + 'd']
                new_coeffs[key] = idwt_axis(L, H, wv, m, ax)

            coeffs = new_coeffs

        return coeffs['']

    return impl


@numba.generated_jit(nopython=True, nogil=True, cache=True)
def promote_level(sizes, dec_lens, level=None):
    have_level = not is_nonelike(level)

    if isinstance(sizes, nbtypes.Integer):
        int_sizes = True
    elif (isinstance(sizes, NUMBA_SEQUENCE_TYPES) and
            isinstance(sizes.dtype, nbtypes.Integer)):
        int_sizes = False
    else:
        raise TypeError("sizes must be an integer or "
                        "sequence of integers")

    if isinstance(dec_lens, nbtypes.Integer):
        int_dec_len = True
    elif (isinstance(dec_lens, NUMBA_SEQUENCE_TYPES) and
            isinstance(dec_lens.dtype, nbtypes.Integer)):
        int_dec_len = False
    else:
        raise TypeError("dec_len must be an integer or "
                        "sequence of integers")

    def impl(sizes, dec_lens, level=None):
        if int_sizes:
            sizes = List([sizes])

        if int_dec_len:
            dec_lens = List([dec_lens])

        max_level = min([dwt_max_level(s, d) for s, d in zip(sizes, dec_lens)])

        if not have_level:
            level = max_level
        elif level < 0:
            raise ValueError("Negative levels are invalid. Minimum level is 0")
        elif level > max_level:
            # with numba.objmode():
            #     warnings.warn("Level value is too high. "
            #                   "All coefficients will experience "
            #                   "boundary effects")
            pass

        return level

    return impl


@numba.generated_jit(nopython=True, nogil=True, fastmath=True, cache=True)
def wavedecn(data, wavelet, mode='zero', level=None, axis=None):
    have_axis = not is_nonelike(axis)

    def impl(data, wavelet, mode='zero', level=None, axis=None):
        if not have_axis:
            axis = List(range(data.ndim))

        paxis = promote_axis(axis, data.ndim)
        naxis = len(paxis)
        # pmodes = promote_mode(mode, naxis)
        pwavelets = [discrete_wavelet(w) for w
                     in promote_wavelets(wavelet, naxis)]
        dec_lens = [w.dec_hi.shape[0] for w in pwavelets]
        sizes = [data.shape[ax] for ax in paxis]
        plevel = promote_level(sizes, dec_lens, level)

        coeffs_list = List()

        a = data

        for i in range(plevel):
            coeffs = dwt(a, wavelet, mode, paxis)
            a = not_optional(coeffs.pop('a' * naxis))
            coeffs_list.append(coeffs)

        coeffs_list.append({'aa': a})

        coeffs_list.reverse()

        return coeffs_list

    return impl


@numba.generated_jit(nopython=True, nogil=True, fastmath=True, cache=True)
def waverecn(coeffs, wavelet, mode='zero', axis=None):
    # ca = coeffs[0]['aa']
    # if not isinstance(ca, nbtypes.npytypes.Array):
    #     raise TypeError("ca must be an ndarray")

    have_axis = not is_nonelike(axis)
    # ndim_slices = (slice(None),) * ca.ndim

    def impl(coeffs, wavelet, mode='zero', axis=None):
        ca = coeffs[0]['aa']
        if len(coeffs) == 1:
            return ca

        coeff_ndims = [ca.ndim]
        coeff_shapes = [ca.shape]

        for c in coeffs[1:]:
            coeff_ndims.extend([v.ndim for v in c.values()])
            coeff_shapes.extend([v.shape for v in c.values()])

        unique_coeff_ndims = np.unique(np.array(coeff_ndims))

        if len(unique_coeff_ndims) == 1:
            ndim = unique_coeff_ndims[0]
        else:
            raise ValueError("Coefficient dimensions don't match")

        if not have_axis:
            axis = List(range(ndim))

        paxes = promote_axis(axis, ndim)
        naxis = len(paxes)

        for idx, c in enumerate(coeffs[1:]):
            c[not_optional('a' * naxis)] = ca
            ca = idwt(c, wavelet, mode, axis)

        return ca

    return impl


@numba.njit(nogil=True, fastmath=True, cache=True)
def ravel_coeffs(coeffs):
    a_coeffs = coeffs[0]['aa']

    ndim = a_coeffs.ndim

    # initialize with the approximation coefficients.
    a_size = a_coeffs.size
    arr_size = a_size
    for c in coeffs[1:]:
        for k, v in c.items():
            arr_size += v.size
    coeff_arr = np.empty((arr_size, ), dtype=a_coeffs.dtype)

    a_slice = slice(a_size)
    coeff_arr[a_slice] = a_coeffs.ravel()

    # initialize list of coefficient slices
    coeff_tuples = List()
    coeff_shapes = List()
    tmp1 = Dict()
    tmp1['aa'] = (0, a_size)
    coeff_tuples.append(tmp1)
    tmp2 = Dict()
    tmp2['aa'] = a_coeffs.shape
    coeff_shapes.append(tmp2)
    coeff_shapes[-1]['aa'] = a_coeffs.shape


    if len(coeffs) == 1:
        return coeff_arr, coeff_tuples, coeff_shapes

    # loop over the detail cofficients, embedding them in coeff_arr
    offset = a_size
    for coeff_dict in coeffs[1:]:
        # new dictionaries for detail coefficient slices and shapes
        tmp1 = Dict()
        tmp2 = Dict()

        # sort to make sure key order is consistent across Python versions
        keys = sorted(coeff_dict.keys())
        for i, key in enumerate(keys):
            d = coeff_dict[key]
            tmp1[key] = (offset, offset + d.size)
            coeff_arr[offset:offset + d.size] = d.ravel()
            tmp2[key] = d.shape
            offset += d.size

        coeff_tuples.append(tmp1)
        coeff_shapes.append(tmp2)

    return coeff_arr, coeff_tuples, coeff_shapes


@numba.njit(nogil=True, fastmath=True, cache=True)
def unravel_coeffs(arr, coeff_tuples, coeff_shapes, output_format='wavedecn'):
    arr = np.asarray(arr)
    coeffs = List()
    tmp = Dict()
    start, stop = coeff_tuples[0]['aa']
    tmp['aa'] = arr[start:stop].reshape(coeff_shapes[0]['aa'])
    coeffs.append(tmp)

    # difference coefficients at each level
    for n in range(1, len(coeff_tuples)):
        shape_dict = coeff_shapes[n]
        d = Dict()
        for k, v in coeff_tuples[n].items():
            start, stop = v
            d[k] = arr[start:stop].reshape(shape_dict[k])
        coeffs.append(d)
    return coeffs


@numba.njit(nogil=True, fastmath=True, cache=True)
def wavelet_setup(x, bases, nlevels):
    # set up dictionary info
    tys = Dict()
    sys = Dict()
    nmax = x[0].ravel().size
    ntot = List()
    ntot.append(nmax)
    for base in bases:
        if base == 'self':
            continue
        alpha = wavedecn(x[0], base, mode='zero', level=nlevels)
        y, ty, sy = ravel_coeffs(alpha)
        tys[base] = ty
        sys[base] = sy

        ntot.append(y.size)
        nmax = np.maximum(nmax, y.size)

    return tys, sys, ntot, nmax



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('DEGRID')


from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.degrid["inputs"].keys():
    defaults[key] = schema.degrid["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.degrid)
def degrid(**kw):
    '''
    Predict model visibilities to measurement sets.
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'degrid_{timestamp}.log')

    from daskms.fsspec_store import DaskMSStore
    msstore = DaskMSStore(opts.ms.rstrip('/'))
    ms = msstore.fs.glob(opts.ms.rstrip('/'))
    try:
        assert len(ms) > 0
        opts.ms = list(map(msstore.fs.unstrip_protocol, ms))
    except:
        raise ValueError(f"No MS at {opts.ms}")

    OmegaConf.set_struct(opts, True)

    if opts.product.upper() not in ["I"]:
                                    # , "Q", "U", "V", "XX", "YX", "XY",
                                    # "YY", "RR", "RL", "LR", "LL"]:
        raise NotImplementedError(f"Product {opts.product} not yet supported")

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _degrid(**opts)

def _degrid(**kw):
    opts = OmegaConf.create(kw)
    from omegaconf import ListConfig
    if not isinstance(opts.ms, list) and not isinstance(opts.ms, ListConfig) :
        opts.ms = [opts.ms]
    OmegaConf.set_struct(opts, True)

    import numpy as np
    from pfb.utils.misc import construct_mappings
    import dask
    from dask.distributed import performance_report
    from dask.graph_manipulation import clone
    from daskms.experimental.zarr import xds_from_zarr
    from daskms import xds_from_storage_ms as xds_from_ms
    from daskms import xds_from_storage_table as xds_from_table
    from daskms import xds_to_storage_table as xds_to_table
    from daskms.optimisation import inlined_array
    import dask.array as da
    from africanus.constants import c as lightspeed
    from africanus.gridding.wgridder.dask import model as im2vis
    from pfb.operators.gridder import comps2vis, _comps2vis_impl
    from pfb.utils.fits import load_fits, data_from_header
    from astropy.io import fits
    from pfb.utils.misc import compute_context
    import xarray as xr
    import sympy as sm
    from sympy.utilities.lambdify import lambdify
    from sympy.parsing.sympy_parser import parse_expr

    mds = xds_from_zarr(opts.mds)[0]

    # grid spec
    cell_rad = mds.cell_rad_x
    cell_deg = np.rad2deg(cell_rad)
    nx = mds.npix_x
    ny = mds.npix_y
    x0 = mds.center_x
    y0 = mds.center_y
    radec = (mds.ra, mds.dec)

    # model func
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(mds.params.values))
    symexpr = parse_expr(mds.parametrisation)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(mds.texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(mds.fexpr)
    ffunc = lambdify(params[1], fexpr)


    # signature (t, f, *params) with
    # t = utime/ref_time
    # f = freq/ref_freq


    group_by = ['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER']

    print('Constructing mapping', file=log)
    row_mapping, freq_mapping, time_mapping, \
        freqs, utimes, ms_chunks, gain_chunks, radecs, \
        chan_widths, uv_max, antpos, poltype = \
            construct_mappings(opts.ms,
                               None,
                               ipi=opts.integrations_per_image,
                               cpi=opts.channels_per_image)


    print("Computing model visibilities", file=log)
    writes = []
    # avoid reading these more than once
    coeffs = mds.coefficients.values
    locx = mds.location_x.values
    locy = mds.location_y.values
    for ms in opts.ms:
        xds = xds_from_ms(ms,
                          chunks=ms_chunks[ms],
                          group_cols=group_by)

        out_data = []
        for ds in xds:
            # TODO - rephase if fields don't match
            # radec = radecs[ms][idt]
            # if not np.array_equal(radec, field.PHASE_DIR.data.squeeze()):
            #     continue
            fid = ds.FIELD_ID
            ddid = ds.DATA_DESC_ID
            scanid = ds.SCAN_NUMBER
            idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"

            # time <-> row mapping
            utime = da.from_array(utimes[ms][idt],
                                  chunks=opts.integrations_per_image)
            tidx = da.from_array(time_mapping[ms][idt]['start_indices'],
                                 chunks=1)
            tcnts = da.from_array(time_mapping[ms][idt]['counts'],
                                  chunks=1)

            ridx = da.from_array(row_mapping[ms][idt]['start_indices'],
                                 chunks=opts.integrations_per_image)
            rcnts = da.from_array(row_mapping[ms][idt]['counts'],
                                  chunks=opts.integrations_per_image)

            # freq <-> band mapping
            freq = da.from_array(freqs[ms][idt],
                                 chunks=opts.channels_per_image)
            fidx = da.from_array(freq_mapping[ms][idt]['start_indices'],
                                 chunks=1)
            fcnts = da.from_array(freq_mapping[ms][idt]['counts'],
                                  chunks=1)

            # number of chunks need to math in mapping and coord
            ntime_out = len(tidx.chunks[0])
            assert len(utime.chunks[0]) == ntime_out
            nfreq_out = len(fidx.chunks[0])
            assert len(freq.chunks[0]) == nfreq_out
            # and they need to match the number of row chunks
            uvw = ds.UVW.data
            assert len(uvw.chunks[0]) == len(tidx.chunks[0])

            vis = comps2vis(uvw,
                            utime,
                            freq,
                            ridx, rcnts,
                            tidx, tcnts,
                            fidx, fcnts,
                            coeffs,
                            locx, locy,
                            modelf,
                            tfunc,
                            ffunc,
                            nx, ny,
                            cell_rad, cell_rad,
                            x0=x0, y0=y0,
                            nthreads=opts.nvthreads,
                            epsilon=opts.epsilon,
                            wstack=opts.wstack)

            # convert to single precision to write to MS
            vis = vis.astype(np.complex64)

            if opts.accumulate:
                vis += getattr(ds, opts.model_column).data

            vis = inlined_array(vis, [uvw])

            out_ds = ds.assign(**{opts.model_column:
                                 (("row", "chan", "corr"), vis)})
            out_data.append(out_ds)

        writes.append(xds_to_table(out_data, ms,
                                   columns=[opts.model_column],
                                   rechunk=True))

    # dask.visualize(writes, color="order", cmap="autumn",
    #                node_attr={"penwidth": "4"},
    #                filename=opts.output_filename + '_degrid_writes_I_ordered_graph.pdf',
    #                optimize_graph=False)
    # dask.visualize(writes, filename=opts.output_filename +
    #                '_degrid_writes_I_graph.pdf', optimize_graph=False)

    with compute_context(opts.scheduler, opts.mds+'_degrid'):
        dask.compute(writes, optimize_graph=False)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here.", file=log)




import numpy as np
from katbeam import JimBeam
import dask.array as da
from pfb.utils.beam import interp_beam, eval_beam
import pytest

pmp = pytest.mark.parametrize

@pmp('beam_model', ('kbl', 'kbuhf', None))
@pmp('nx', (24, 128))
@pmp('ny', (64, 92))
def test_beam(beam_model, nx, ny):
    freq = 1e3
    cell_deg = 1e-2
    beam = interp_beam(freq, nx, ny, cell_deg, beam_model)

    l = (-(nx//2) + da.arange(nx)) * cell_deg
    m = (-(ny//2) + da.arange(ny)) * cell_deg
    ll, mm = da.meshgrid(l, m, indexing='ij')

    bvals = eval_beam(beam, ll, mm).compute()

    if beam_model is None:
        assert (bvals==1).all()
    elif beam_model == 'kbl':
        jb = JimBeam('MKAT-AA-L-JIM-2020').I
        assert (bvals == jb(ll, mm, freq)).all()
    elif beam_model == 'kbuhf':
        jb = JimBeam('MKAT-AA-UHF-JIM-2020').I
        assert (bvals == jb(ll, mm, freq)).all()



import pytest
from pathlib import Path
import requests
import tarfile

test_root_path = Path(__file__).resolve().parent
test_data_path = Path(test_root_path, "data")
test_data_path.mkdir(parents=True, exist_ok=True)

_data_tar_name = "test_ascii_1h60.0s.MS.tar.gz"
_ms_name = "test_ascii_1h60.0s.MS"

data_tar_path = Path(test_data_path, _data_tar_name)
ms_path = Path(test_data_path, _ms_name)

#https://drive.google.com/file/d/1rfGXGjjJ2XtF26LImlyJzCJMCNQZgEFT/view?usp=sharing

gdrive_id = "1rfGXGjjJ2XtF26LImlyJzCJMCNQZgEFT"

url = "https://drive.google.com/uc?id={id}".format(id=gdrive_id)

def pytest_sessionstart(session):
    """Called after Session object has been created, before run test loop."""

    if ms_path.exists():
        print("Test data already present - not downloading.")
    else:
        print("Test data not found - downloading...")
        download = requests.get(url)  # , params={"dl": 1}
        with open(data_tar_path, 'wb') as f:
            f.write(download.content)
        with tarfile.open(data_tar_path, "r:gz") as tar:
            tar.extractall(path=test_data_path)
        data_tar_path.unlink()
        print("Test data successfully downloaded.")


@pytest.fixture(scope='session')
def ms_name():
    return str(ms_path)



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
from functools import partial
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('CLEAN')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.clean["inputs"].keys():
    defaults[key] = schema.clean["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.clean)
def clean(**kw):
    '''
    Modified single-scale clean.
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'clean_{timestamp}.log')

    if opts.nworkers is None:
        if opts.scheduler=='distributed':
            opts.nworkers = opts.nband
        else:
            opts.nworkers = 1

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        # numpy imports have to happen after this step
        from pfb import set_client
        set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _clean(**opts)


def _clean(**kw):
    opts = OmegaConf.create(kw)
    # always combine over ds during cleaning
    opts['mean_ds'] = True
    OmegaConf.set_struct(opts, True)

    import numpy as np
    import xarray as xr
    import numexpr as ne
    import dask
    import dask.array as da
    from dask.distributed import performance_report
    from pfb.utils.fits import (set_wcs, save_fits, dds2fits,
                                dds2fits_mfs, load_fits)
    from pfb.utils.misc import dds2cubes
    from pfb.deconv.hogbom import hogbom
    from pfb.deconv.clark import clark
    from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
    from pfb.opt.pcg import pcg, pcg_psf
    from pfb.operators.hessian import hessian_xds, hessian
    from scipy import ndimage
    from copy import copy

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    dds_name = f'{basename}_{opts.postfix}.dds'
    dds = xds_from_zarr(dds_name, chunks={'row':-1,
                                          'chan':-1,
                                          'x':-1,
                                          'y':-1,
                                          'x_psf':-1,
                                          'y_psf':-1,
                                          'yo2':-1})
    if opts.memory_greedy:
        dds = dask.persist(dds)[0]

    nx_psf, ny_psf = dds[0].x_psf.size, dds[0].y_psf.size
    lastsize = ny_psf

    # stitch dirty/psf in apparent scale
    output_type = dds[0].DIRTY.dtype
    dirty, model, residual, psf, psfhat, _, wsums, _ = dds2cubes(
                                                            dds,
                                                            opts.nband,
                                                            apparent=True)
    # because fuck dask
    model = np.require(model, requirements='CAW')


    wsum = np.sum(wsums)
    psf_mfs = np.sum(psf, axis=0)
    # This is onlt the case for a psf at (l=0,m=0)
    # assert (psf_mfs.max() - 1.0) < 2*opts.epsilon
    dirty_mfs = np.sum(dirty, axis=0)
    if residual is None:
        residual = dirty.copy()
        residual_mfs = dirty_mfs.copy()
    else:
        residual_mfs = np.sum(residual, axis=0)

    # for intermediary results (not currently written)
    freq_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
    freq_out = np.unique(np.array(freq_out))
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    x0 = dds[0].x0
    y0 = dds[0].y0
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)

    # TODO - check coordinates match
    # Add option to interp onto coordinates?
    if opts.mask is not None:
        mask = load_fits(mask, dtype=output_type).squeeze()
        assert mask.shape == (nx, ny)
        mask = mask.astype(output_type)
        print('Using provided fits mask', file=log)
    else:
        mask = np.ones((nx, ny), dtype=output_type)

    # set up vis space Hessian
    hessopts = {}
    hessopts['cell'] = dds[0].cell_rad
    hessopts['wstack'] = opts.wstack
    hessopts['epsilon'] = opts.epsilon
    hessopts['double_accum'] = opts.double_accum
    hessopts['nthreads'] = opts.nvthreads
    hessopts['x0'] = x0
    hessopts['y0'] = y0
    # always clean in apparent scale so no beam
    # mask is applied to residual after hessian application
    hess = partial(hessian_xds, xds=dds, hessopts=hessopts,
                   wsum=wsum, sigmainv=0,
                   mask=np.ones((nx, ny), dtype=output_type),
                   compute=True, use_beam=False)

    # PCG related options for flux mop
    cgopts = {}
    cgopts['tol'] = opts.cg_tol
    cgopts['maxit'] = opts.cg_maxit
    cgopts['minit'] = opts.cg_minit
    cgopts['verbosity'] = opts.cg_verbose
    cgopts['report_freq'] = opts.cg_report_freq
    cgopts['backtrack'] = opts.backtrack


    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()

    if opts.threshold is None:
        threshold = opts.sigmathreshold * rms
    else:
        threshold = opts.threshold

    print(f"Iter 0: peak residual = {rmax:.3e}, rms = {rms:.3e}",
          file=log)
    for k in range(opts.nmiter):
        print("Cleaning", file=log)
        x, status = clark(mask*residual, psf, psfhat, wsums/wsum,
                          threshold=threshold,
                          gamma=opts.gamma,
                          pf=opts.peak_factor,
                          maxit=opts.minor_maxit,
                          subpf=opts.sub_peak_factor,
                          submaxit=opts.subminor_maxit,
                          verbosity=opts.verbose,
                          report_freq=opts.report_freq,
                          sigmathreshold=opts.sigmathreshold,
                          nthreads=opts.nvthreads)
        model += x

        print("Getting residual", file=log)
        convimage = hess(model)
        ne.evaluate('dirty - convimage', out=residual,
                    casting='same_kind')
        ne.evaluate('sum(residual, axis=0)', out=residual_mfs,
                    casting='same_kind')

        # report rms where there aren;t any model components
        tmp_mask = ~np.any(model, axis=0)
        rms = np.std(residual_mfs[tmp_mask])
        rmax = np.abs(residual_mfs).max()

        if opts.threshold is None:
            threshold = opts.sigmathreshold * rms
        else:
            threshold = opts.threshold

        # trigger flux mop if clean has stalled, not converged or
        # we have reached the final iteration/threshold
        status |= k == opts.nmiter-1
        status |= rmax <= threshold
        if opts.mop_flux and status:
            print(f"Mopping flux at iter {k+1}", file=log)
            mopmask = np.any(model, axis=0)
            if opts.dirosion:
                struct = ndimage.generate_binary_structure(2, opts.dirosion)
                mopmask = ndimage.binary_dilation(mopmask, structure=struct)
                mopmask = ndimage.binary_erosion(mopmask, structure=struct)
            x0 = np.zeros_like(x)
            x0[:, mopmask] = residual_mfs[mopmask]
            # TODO - applying mask as beam is wasteful
            mopmask = mopmask[None, :, :].astype(residual.dtype)
            x = pcg_psf(psfhat,
                        mopmask*residual,
                        x0,
                        mopmask,
                        lastsize,
                        opts.nvthreads,
                        rmax,  # used as sigmainv
                        cgopts)

            model += opts.mop_gamma*x

            save_fits(residual_mfs,
                      f'{basename}_{opts.postfix}_premop{k}_resid_mfs.fits',
                      hdr_mfs)

            print("Getting residual", file=log)
            convimage = hess(model)
            ne.evaluate('dirty - convimage', out=residual,
                        casting='same_kind')
            ne.evaluate('sum(residual, axis=0)', out=residual_mfs,
                        casting='same_kind')

            save_fits(residual_mfs,
                      f'{basename}_{opts.postfix}_postmop{k}_resid_mfs.fits',
                      hdr_mfs)

            tmp_mask = ~np.any(model, axis=0)
            rms = np.std(residual_mfs[tmp_mask])
            rmax = np.abs(residual_mfs).max()

            if opts.threshold is None:
                threshold = opts.sigmathreshold * rms
            else:
                threshold = opts.threshold

        print(f"Iter {k+1}: peak residual = {rmax:.3e}, rms = {rms:.3e}",
              file=log)

        print("Updating results", file=log)
        dds_out = []
        for ds in dds:
            b = ds.bandid
            r = da.from_array(residual[b]*wsum)
            m = da.from_array(model[b])
            ds_out = ds.assign(**{'RESIDUAL': (('x', 'y'), r),
                                  'MODEL': (('x', 'y'), m)})
            dds_out.append(ds_out)
        writes = xds_to_zarr(dds_out, dds_name,
                             columns=('RESIDUAL', 'MODEL'), rechunk=True)
        dask.compute(writes)

        if rmax <= threshold:
            print("Terminating because final threshold has been reached",
                  file=log)
            break

    dds = xds_from_zarr(dds_name, chunks={'x': -1, 'y': -1})

    # convert to fits files
    fitsout = []
    if opts.fits_mfs:
        fitsout.append(dds2fits_mfs(dds, 'RESIDUAL', f'{basename}_{opts.postfix}', norm_wsum=True))
        fitsout.append(dds2fits_mfs(dds, 'MODEL', f'{basename}_{opts.postfix}', norm_wsum=False))

    if opts.fits_cubes:
        fitsout.append(dds2fits(dds, 'RESIDUAL', f'{basename}_{opts.postfix}', norm_wsum=True))
        fitsout.append(dds2fits(dds, 'MODEL', f'{basename}_{opts.postfix}', norm_wsum=False))

    if len(fitsout):
        print("Writing fits", file=log)
        dask.compute(fitsout)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here.", file=log)



"""
Pre-conditioned Forward Backward Clean algorithm

author - Landman Bester
email  - lbester@ska.ac.za
date   - 31/03/2020
"""
__version__ = '0.0.1'

import os

def set_client(opts, stack, log, scheduler='distributed'):

    from omegaconf import open_dict
    # attempt somewhat intelligent default setup
    import multiprocessing
    nthreads_max = max(multiprocessing.cpu_count(), 1)
    if opts.nvthreads is None:
        # we allocate half by default
        nthreads_tot = max(nthreads_max//2, 1)
        if opts.scheduler in ['single-threaded', 'sync']:
            with open_dict(opts):
                opts.nvthreads = nthreads_tot
        else:
            ndask_chunks = opts.nthreads_dask*opts.nworkers
            nvthreads = max(nthreads_tot//ndask_chunks, 1)
            with open_dict(opts):
                opts.nvthreads = nvthreads

    os.environ["OMP_NUM_THREADS"] = str(opts.nvthreads)
    os.environ["OPENBLAS_NUM_THREADS"] = str(opts.nvthreads)
    os.environ["MKL_NUM_THREADS"] = str(opts.nvthreads)
    os.environ["VECLIB_MAXIMUM_THREADS"] = str(opts.nvthreads)
    import numexpr as ne
    max_cores = ne.detect_number_of_cores()
    # ne_threads = min(max_cores, opts.nvthreads)
    os.environ["NUMEXPR_NUM_THREADS"] = str(max_cores)
    os.environ["NUMBA_NUM_THREADS"] = str(max_cores)

    import dask
    if scheduler=='distributed':
        # TODO - investigate what difference this makes
        # with dask.config.set({"distributed.scheduler.worker-saturation":  1.1}):
        #     client = distributed.Client()
        # set up client
        if opts.host_address is not None:
            from distributed import Client
            print("Initialising distributed client.", file=log)
            client = stack.enter_context(Client(opts.host_address))
        else:
            if opts.nthreads_dask * opts.nvthreads > nthreads_max:
                print("Warning - you are attempting to use more threads than "
                      "available. This may lead to suboptimal performance.",
                      file=log)
            from dask.distributed import Client, LocalCluster
            print("Initialising client with LocalCluster.", file=log)
            with dask.config.set({"distributed.scheduler.worker-saturation":  1.1}):
                cluster = LocalCluster(processes=True, n_workers=opts.nworkers,
                                    threads_per_worker=opts.nthreads_dask,
                                    memory_limit=0)  # str(mem_limit/nworkers)+'GB'
                cluster = stack.enter_context(cluster)
                client = stack.enter_context(Client(cluster))

        from quartical.scheduling import install_plugin
        client.run_on_scheduler(install_plugin)
        client.wait_for_workers(opts.nworkers)
    elif scheduler in ['sync', 'single-threaded']:
        dask.config.set(scheduler=scheduler)
        print(f"Initialising with synchronous scheduler",
              file=log)
    elif scheduler=='threads':
        from multiprocessing.pool import ThreadPool
        dask.config.set(pool=ThreadPool(opts.nthreads_dask))
        print(f"Initialising ThreadPool with {opts.nthreads_dask} threads",
              file=log)
    else:
        raise ValueError(f"Unknown scheduler option {opts.scheduler}")

    # return updated opts
    return opts


def logo():
    print("""
    ███████████  ███████████ ███████████
   ░░███░░░░░███░░███░░░░░░█░░███░░░░░███
    ░███    ░███ ░███   █ ░  ░███    ░███
    ░██████████  ░███████    ░██████████
    ░███░░░░░░   ░███░░░█    ░███░░░░░███
    ░███         ░███  ░     ░███    ░███
    █████        █████       ███████████
   ░░░░░        ░░░░░       ░░░░░░░░░░░
    """)



from itertools import product
import numba
from numba.cpython.unsafe.tuple import tuple_setitem
import numpy as np
import pywt
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pytest

from pfb.wavelets.wavelets import (dwt, dwt_axis,
                                   idwt, idwt_axis,
                                   dwt_max_level,
                                   str_to_int,
                                   coeff_product,
                                   promote_axis,
                                   promote_level,
                                   discrete_wavelet,
                                   wavedecn, waverecn,
                                   ravel_coeffs, unravel_coeffs)
from pfb.wavelets.modes import (Modes,
                                promote_mode,
                                mode_str_to_enum)

from pfb.wavelets.intrinsics import slice_axis


convert_mode = numba.njit(lambda s: mode_str_to_enum(s))


@pytest.mark.parametrize("ndim", [1, 2, 3])
@pytest.mark.parametrize("extent", [-1, 0, 1])
def test_slice_axis(ndim, extent):
    @numba.njit
    def fn(a, index, axis=1, extent=None):
        return slice_axis(a, index, axis, extent)

    A = np.random.random(np.random.randint(4, 10, size=ndim))
    assert A.ndim == ndim

    for axis in range(A.ndim):
        # Randomly choose indexes within the array
        tup_idx = tuple(np.random.randint(0, d) for d in A.shape)
        # Replace index with slice along desired axis
        ax_size = A.shape[axis]

        ext = None if extent == 0 else ax_size + extent
        slice_idx = tuple(slice(ext) if a == axis
                          else i for a, i in enumerate(tup_idx))

        As = A[slice_idx]
        B = fn(A, tup_idx, axis, ext)

        assert_array_equal(As, B)

        if ndim == 1:
            assert B.flags.c_contiguous is As.flags.c_contiguous
            assert B.flags.f_contiguous is As.flags.f_contiguous
            assert B.flags.aligned is As.flags.aligned
            assert B.flags.writeable is As.flags.writeable
            assert B.flags.writebackifcopy is As.flags.writebackifcopy
            # TODO - is this harmless?
            # assert B.flags.updateifcopy is As.flags.updateifcopy

            # TODO(sjperkins)
            # Why is owndata True in the
            # case of the numba intrinsic, but
            # not in the case of numpy?
            assert B.flags.owndata is not (extent or As.flags.owndata)
        else:
            assert B.flags == As.flags

        # Check that modifying the numba slice
        # modifies the numpy slice
        B[:] = np.arange(B.shape[0])
        assert_array_equal(As, B)


def test_internal_slice_axis():
    @numba.njit
    def fn(A):
        for axis in range(A.ndim):
            for i in np.ndindex(*tuple_setitem(A.shape, axis, 1)):
                S = slice_axis(A, i, axis, None)

                if S.flags.c_contiguous != (S.itemsize == S.strides[0]):
                    raise ValueError("contiguity flag doesn't match layout")

    fn(np.random.random((8, 9, 10)))


@pytest.mark.parametrize("repeat", range(5))
def test_coeff_product(repeat):
    res = coeff_product('ad', repeat=repeat)
    coeffs = [''.join(c) for c in product('ad', repeat=repeat)]
    assert list(res) == coeffs


def test_str_to_int():
    assert str_to_int("111") == 111
    assert str_to_int("23") == 23
    assert str_to_int("3") == 3


# def test_promote_mode():
#     assert [Modes.symmetric] == list(promote_mode("zero", 1))
#     assert [Modes.symmetric]*3 == list(promote_mode("zero", 3))

#     assert [Modes.symmetric] == list(promote_mode(["zero"], 1))
#     assert [Modes.symmetric] == list(promote_mode(("zero",), 1))

#     with pytest.raises(ValueError):
#         assert [Modes.symmetric] == list(promote_mode(["zero"], 2))

#     list_inputs = ["zero"]
#     tuple_inputs = tuple(list_inputs)
#     result_enums = [Modes.zero_pad]

#     assert result_enums == list(promote_mode(list_inputs, 2))
#     assert result_enums == list(promote_mode(tuple_inputs, 2))

#     with pytest.raises(ValueError):
#         assert result_enums == list(promote_mode(list_inputs, 3))

#     with pytest.raises(ValueError):
#         assert result_enums == list(promote_mode(list_inputs, 1))


def test_promote_axis():
    assert [0] == list(promote_axis(0, 1))
    assert [0] == list(promote_axis([0], 1))
    assert [0] == list(promote_axis((0,), 1))

    with pytest.raises(ValueError):
        assert [0, 1] == list(promote_axis((0, 1), 1))

    assert [0, 1] == list(promote_axis((0, 1), 2))
    assert [0, 1] == list(promote_axis([0, 1], 2))

    assert [0, 1] == list(promote_axis((0, 1), 3))


@pytest.mark.parametrize("data", [500, 100, 12])
@pytest.mark.parametrize("filter", [500, 100, 24])
def test_dwt_max_level(data, filter):
    pywt = pytest.importorskip("pywt")
    assert pywt.dwt_max_level(data, filter) == dwt_max_level(data, filter)


@pytest.mark.parametrize("wavelet", ["db1", "db4", "db5"])
def test_discrete_wavelet(wavelet):
    pfb_wave = discrete_wavelet(wavelet)

    pywt = pytest.importorskip("pywt")
    py_wave = pywt.Wavelet(wavelet)

    # assert py_wave.support_width == pfb_wave.support_width
    assert py_wave.orthogonal == pfb_wave.orthogonal
    assert py_wave.biorthogonal == pfb_wave.biorthogonal
    #assert py_wave.compact_support == pfb_wave.compact_support
    assert py_wave.family_name == pfb_wave.family_name
    assert py_wave.short_family_name == pfb_wave.short_name
    assert py_wave.vanishing_moments_phi == pfb_wave.vanishing_moments_phi
    assert py_wave.vanishing_moments_psi == pfb_wave.vanishing_moments_psi

    assert_array_almost_equal(py_wave.rec_lo, pfb_wave.rec_lo)
    assert_array_almost_equal(py_wave.dec_lo, pfb_wave.dec_lo)
    assert_array_almost_equal(py_wave.rec_hi, pfb_wave.rec_hi)
    assert_array_almost_equal(py_wave.dec_hi, pfb_wave.dec_hi)


@pytest.mark.parametrize("wavelet", ["db1", "db4", "db5"])
@pytest.mark.parametrize("data_shape", [(13,), (12, 7)])
@pytest.mark.parametrize("mode", ["zero"])
def test_dwt_idwt_axis(wavelet, mode, data_shape):
    pywt = pytest.importorskip("pywt")
    data = np.random.random(size=data_shape)
    enum_mode = convert_mode(mode)

    pywt_dwt_axis = pywt._dwt.dwt_axis
    pywt_idwt_axis = pywt._dwt.idwt_axis

    pywt_wavelet = pywt.Wavelet(wavelet)
    pywt_mode = pywt.Modes.from_object(mode)

    wavelet = discrete_wavelet(wavelet)

    for axis in reversed(range(len(data_shape))):
        # Deconstruct
        ca, cd = dwt_axis(data, wavelet, enum_mode, axis)
        pywt_ca, pywt_cd = pywt_dwt_axis(data, pywt_wavelet, pywt_mode, axis)
        assert_array_almost_equal(ca, pywt_ca)
        assert_array_almost_equal(cd, pywt_cd)

        # Reconstruct with both approximation and detail
        pywt_out = pywt_idwt_axis(ca, cd, pywt_wavelet, pywt_mode, axis)
        output = idwt_axis(ca, cd, wavelet, enum_mode, axis)
        assert_array_almost_equal(output, pywt_out)

        # Reconstruct with approximation only
        pywt_out = pywt_idwt_axis(ca, None, pywt_wavelet, pywt_mode, axis)
        output = idwt_axis(ca, None, wavelet, enum_mode, axis)
        assert_array_almost_equal(output, pywt_out)

        # Reconstruct with detail only
        pywt_out = pywt_idwt_axis(None, cd, pywt_wavelet, pywt_mode, axis)
        output = idwt_axis(None, cd, wavelet, enum_mode, axis)
        assert_array_almost_equal(output, pywt_out)


def test_dwt_idwt():
    pywt = pytest.importorskip("pywt")
    data = np.random.random((5, 8, 11))

    res = dwt(data, "db1", "zero")
    pywt_res = pywt.dwtn(data, "db1", "zero")
    for k, v in res.items():
        assert_array_almost_equal(v, pywt_res[k])

    res = dwt(data, "db1", "zero", 1)
    pywt_res = pywt.dwtn(data, "db1", "zero", (1,))
    for k, v in res.items():
        assert_array_almost_equal(v, pywt_res[k])

    res = dwt(data, ("db1", "db2"), ("zero", "zero"), (0, 1))
    pywt_res = pywt.dwtn(data, ("db1", "db2"), ("zero", "zero"), (0, 1))
    for k, v in res.items():
        assert_array_almost_equal(v, pywt_res[k])

    output = idwt(res, ("db1", "db2"), ("zero", "zero"), (0, 1))
    pywt_out = pywt.idwtn(pywt_res, ("db1", "db2"), ("zero", "zero"), (0, 1))
    assert_array_almost_equal(output, pywt_out)


@pytest.mark.parametrize("data_shape", [(64, 128)])
@pytest.mark.parametrize("complex_data", [True, False])
@pytest.mark.parametrize("level", list(range(10)))
@pytest.mark.parametrize("mode", ["zero"])
@pytest.mark.parametrize("wavelet", ["db1", "db4", "db5"])
def test_wavedecn_waverecn(data_shape, wavelet, mode, level, complex_data):
    pywt = pytest.importorskip("pywt")
    data = np.random.random(data_shape)

    if complex_data:
        data = data + np.random.random(data_shape) * 1j

    out = pywt.wavedecn(data, wavelet, mode)
    coeffs = wavedecn(data, wavelet, mode)

    assert_array_almost_equal(coeffs[0]['aa'], out[0])

    for d1, d2 in zip(out[1:], coeffs[1:]):
        assert list(d1.keys()) == list(d2.keys())

        for k, v in d1.items():
            assert_array_almost_equal(v, d2[k])

    pywt_rec = pywt.waverecn(out, wavelet, mode)
    rec = waverecn(coeffs, wavelet, mode)
    assert_array_almost_equal(pywt_rec, rec)

    # out = pywt.wavedecn(data, wavelet, mode, axes=(1, 2))
    # coeffs = wavedecn(data, wavelet, mode, axis=(1, 2))

    # assert_array_almost_equal(coeffs[0]['aa'], out[0])

    # for d1, d2 in zip(out[1:], coeffs[1:]):
    #     assert list(d1.keys()) == list(d2.keys())

    #     for k, v in d1.items():
    #         assert_array_almost_equal(v, d2[k])

    # pywt_rec = pywt.waverecn(out, wavelet, mode, axes=(1, 2))
    # rec = waverecn(coeffs, wavelet, mode, axis=(1, 2))
    # assert_array_almost_equal(pywt_rec, rec)

    # # Test various levels of decomposition
    # out = pywt.wavedecn(data, wavelet, mode, level=level, axes=(1, 2))
    # coeffs = wavedecn(data, wavelet, mode, level=level, axis=(1, 2))

    # assert_array_almost_equal(coeffs[0]['aa'], out[0])

    # for d1, d2 in zip(out[1:], coeffs[1:]):
    #     assert list(d1.keys()) == list(d2.keys())

    #     for k, v in d1.items():
    #         assert_array_almost_equal(v, d2[k])

    # pywt_rec = pywt.waverecn(out, wavelet, mode, axes=(1, 2))
    # rec = waverecn(coeffs, wavelet, mode, axis=(1, 2))
    # assert_array_almost_equal(pywt_rec, rec)

@pytest.mark.parametrize("nx", (24, 120))
@pytest.mark.parametrize("ny", (68, 125))
@pytest.mark.parametrize("level", list(range(4)))
@pytest.mark.parametrize("mode", ["zero"])
@pytest.mark.parametrize("wavelet", ["db1", "db4", "db5"])
def test_ravel_coeffs(nx, ny, level, mode, wavelet):
    # pywt = pytest.importorskip("pywt")
    from time import time
    x = np.random.randn(nx, ny)

    out = pywt.wavedecn(x, wavelet, mode, level=level)
    z, iy, sy = pywt.ravel_coeffs(out)

    coeffs = wavedecn(x, wavelet, mode, level=level)
    z2, iy2, sy2 = ravel_coeffs(coeffs)

    assert_array_almost_equal(z, z2, decimal=13)

    outrec = pywt.unravel_coeffs(z, iy, sy)
    outrec2 = unravel_coeffs(z2, iy2, sy2)

    assert np.allclose(outrec[0], outrec2[0]['aa'])
    for i in range(1, len(coeffs)):
        assert np.allclose(outrec[i]['ad'], outrec2[i]['ad'])
        assert np.allclose(outrec[i]['da'], outrec2[i]['da'])
        assert np.allclose(outrec[i]['dd'], outrec2[i]['dd'])


# test_wavedecn_waverecn((1024,1024), 'db1', 'zero', 3, False)
# test_ravel_coeffs(2048, 1024, 2, 'zero', 'db3')



import numpy as np
import dask.array as da
from operator import getitem
from distributed import wait, get_client, as_completed
from scipy.linalg import norm
from copy import deepcopy
import pyscilog
log = pyscilog.get_logger('PM')


def power_method(
        A,
        imsize,
        b0=None,
        tol=1e-5,
        maxit=250,
        verbosity=1,
        report_freq=25):
    if b0 is None:
        b = np.random.randn(*imsize)
        b /= norm(b)
    else:
        b = b0 / norm(b0)
    beta = 1.0
    eps = 1.0
    k = 0
    bp = b.copy()
    while eps > tol and k < maxit:
        b = A(bp)
        bnorm = np.linalg.norm(b)
        betap = beta
        beta = np.vdot(bp, b) / np.vdot(bp, bp)
        b /= bnorm
        eps = np.linalg.norm(beta - betap) / betap
        k += 1
        bp[...] = b[...]

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} eps = {eps:.3e}", file=log)

    if k == maxit:
        if verbosity:
            print(f"Maximum iterations reached. "
                  f"eps = {eps:.3e}, beta = {beta:.3e}", file=log)
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations. "
                  f"beta = {beta:.3e}", file=log)
    return beta, b


def power(A, bp, bnorm):
    bp /= bnorm
    b = A(bp)
    bsumsq = np.sum(b**2)
    beta_num = np.vdot(b, bp)
    beta_den = np.vdot(bp, bp)

    return b, bsumsq, beta_num, beta_den

def sumsq(b):
    return np.sum(b**2)

def bnormf(bsumsq):
    return np.sqrt(np.sum(bsumsq))

def betaf(beta_num, beta_den):
    return np.sum(beta_num)/np.sum(beta_den)

def power_method_dist(Afs,
                      nx,
                      ny,
                      nband,
                      tol=1e-5,
                      maxit=200):

    client = get_client()
    names = [w['name'] for w in client.scheduler_info()['workers'].values()]

    b = []
    bssq = []
    bnum = []
    bden = []
    for i, (wid, A) in enumerate(Afs.items()):
        b.append(client.submit(np.random.randn, nx, ny,
                          workers={wid}))
        bssq.append(client.submit(sumsq, b[i],
                             workers={wid}))
        # this just initialises the lists required below
        bnum.append(1)
        bden.append(1)
    # wid corresponds to last worker
    bnorm = client.submit(bnormf, bssq,
                          workers={names[0]}).result()
    beta = 1
    for k in range(maxit):
        for i, (wid, A) in enumerate(Afs.items()):
            fut = client.submit(power, A, b[i], bnorm,
                                workers={wid})

            b[i] = client.submit(getitem, fut, 0, workers={wid})
            bssq[i] = client.submit(getitem, fut, 1, workers={wid})
            bnum[i] = client.submit(getitem, fut, 2, workers={wid})
            bden[i] = client.submit(getitem, fut, 3, workers={wid})

        bnorm = client.submit(bnormf, bssq,
                              workers={wid})
        betap = beta
        beta = client.submit(betaf, bnum, bden,
                             workers={wid}).result()

        eps = np.abs(betap - beta)/betap
        if eps < tol:
            break

    return beta


def power2(A, bp, bnorm):
    bp /= bnorm
    b = A(bp)
    bsumsq = da.sum(b**2)
    beta_num = da.vdot(b, bp)
    beta_den = da.vdot(bp, bp)

    return b, bsumsq, beta_num, beta_den


def power_method_persist(ddsf,
                         Af,
                         nx,
                         ny,
                         nband,
                         tol=1e-5,
                         maxit=200):

    client = get_client()
    b = []
    bssq = []
    for ds in ddsf:
        wid = ds.worker
        tmp = client.persist(da.random.normal(0, 1, (nx, ny)),
                             workers={wid})
        b.append(tmp)
        bssq.append(da.sum(b**2))

    bssq = da.stack(bssq)
    bnorm = da.sqrt(da.sum(bssq))
    bp = deepcopy(b)
    beta_num = [da.array()]

    for k in range(maxit):
        for i, ds, A in enumerate(zip(ddsf, Af)):
            bp[i] = b[i]/bnorm
            b[i] = A(bp[i])
            bssq[i] = da.sum(b[i]**2)
            beta_num = da.vdot(b, bp)
            beta_den = da.vdot(bp, bp)










import pytest
from pathlib import Path
from xarray import Dataset
from collections import namedtuple
import dask
import dask.array as da
from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
pmp = pytest.mark.parametrize


def test_spotless(ms_name):
    '''
    # TODO - currently we just check that this runs through.
    # What should the passing criteria be?
    '''

    import numpy as np
    np.random.seed(420)
    from numpy.testing import assert_allclose
    import dask
    from daskms import xds_from_ms, xds_from_table, xds_to_table
    from pfb.utils.misc import Gaussian2D, give_edges
    from africanus.constants import c as lightspeed
    from ducc0.fft import good_size
    from ducc0.wgridder.experimental import dirty2vis
    from pfb.parser.schemas import schema
    from pfb.workers.init import _init
    from pfb.workers.grid import _grid
    from pfb.workers.spotless import _spotless
    from pfb.workers.model2comps import _model2comps
    from pfb.workers.degrid import _degrid
    import sympy as sm
    from sympy.utilities.lambdify import lambdify
    from sympy.parsing.sympy_parser import parse_expr


    test_dir = Path(ms_name).resolve().parent
    xds = xds_from_ms(ms_name,
                      chunks={'row': -1, 'chan': -1})[0]
    spw = xds_from_table(f'{ms_name}::SPECTRAL_WINDOW')[0]
    # ms = table(str(test_dir / 'test_ascii_1h60.0s.MS'), readonly=False)
    # spw = table(str(test_dir / 'test_ascii_1h60.0s.MS::SPECTRAL_WINDOW'))

    # utime = np.unique(ms.getcol('TIME'))
    utime = np.unique(xds.TIME.values)
    freq = spw.CHAN_FREQ.values.squeeze()
    # freq = spw.getcol('CHAN_FREQ').squeeze()
    freq0 = np.mean(freq)

    ntime = utime.size
    nchan = freq.size
    # nant = np.maximum(ms.getcol('ANTENNA1').max(), ms.getcol('ANTENNA2').max()) + 1
    nant = np.maximum(xds.ANTENNA1.values.max(), xds.ANTENNA2.values.max()) + 1

    # ncorr = ms.getcol('FLAG').shape[-1]
    ncorr = xds.corr.size

    # uvw = ms.getcol('UVW')
    uvw = xds.UVW.values
    nrow = uvw.shape[0]
    u_max = abs(uvw[:, 0]).max()
    v_max = abs(uvw[:, 1]).max()
    uv_max = np.maximum(u_max, v_max)

    # image size
    cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

    srf = 2.0
    cell_rad = cell_N / srf
    cell_deg = cell_rad * 180 / np.pi
    cell_size = cell_deg * 3600
    print("Cell size set to %5.5e arcseconds" % cell_size)

    # the test will fail in intrinsic if sources fall near beam sidelobes
    fov = 1.0
    npix = good_size(int(fov / cell_deg))
    while npix % 2:
        npix += 1
        npix = good_size(npix)

    nx = npix
    ny = npix

    print("Image size set to (%i, %i, %i)" % (nchan, nx, ny))

    # model
    model = np.zeros((nchan, nx, ny), dtype=np.float64)
    nsource = 25
    border = np.maximum(int(0.15*nx), int(0.15*ny))
    Ix = np.random.randint(border, npix-border, nsource)
    Iy = np.random.randint(border, npix-border, nsource)
    alpha = -0.7 + 0.1 * np.random.randn(nsource)
    I0 = 1.0 + np.exp(np.random.randn(nsource))
    extentx = np.random.randint(3, int(0.1*nx), nsource)
    extenty = np.random.randint(3, int(0.1*nx), nsource)
    pas = np.random.random(nsource) * 180
    x = -(nx/2) + np.arange(nx)
    y = -(nx/2) + np.arange(ny)
    xin, yin = np.meshgrid(x, y, indexing='ij')
    for i in range(nsource):
        emaj = np.maximum(extentx[i], extenty[i])
        emin = np.minimum(extentx[i], extenty[i])
        gauss = Gaussian2D(xin, yin, GaussPar=(emaj, emin, pas[i]))
        mx, my, gx, gy = give_edges(Ix[i], Iy[i], nx, ny, nx, ny)
        spectrum = I0[i] * (freq/freq0) ** alpha[i]
        model[:, mx, my] += spectrum[:, None, None] * gauss[None, gx, gy]

    # model vis
    epsilon = 1e-7
    model_vis = np.zeros((nrow, nchan, ncorr), dtype=np.complex128)
    for c in range(nchan):
        model_vis[:, c:c+1, 0] = dirty2vis(uvw=uvw,
                                           freq=freq[c:c+1],
                                           dirty=model[c],
                                           pixsize_x=cell_rad,
                                           pixsize_y=cell_rad,
                                           epsilon=epsilon,
                                           do_wgridding=True,
                                           divide_by_n=False,
                                           flip_v=False,
                                           nthreads=8,
                                           sigma_min=1.1,
                                           sigma_max=3.0)
        model_vis[:, c, -1] = model_vis[:, c, 0]

    model_vis += (np.random.randn(nrow, nchan, ncorr) +
                  1.0j*np.random.randn(nrow, nchan, ncorr))

    model_vis = da.from_array(model_vis, chunks=(-1,-1,-1))
    xds['DATA'] = (('row','chan','coor'), model_vis)
    writes = [xds_to_table(xds, ms_name, columns='DATA')]
    dask.compute(writes)

    # set defaults from schema
    init_args = {}
    for key in schema.init["inputs"].keys():
        init_args[key] = schema.init["inputs"][key]["default"]
    # overwrite defaults
    outname = str(test_dir / 'test')
    init_args["ms"] = str(test_dir / 'test_ascii_1h60.0s.MS')
    init_args["output_filename"] = outname
    init_args["data_column"] = "DATA"
    # init_args["weight_column"] = 'WEIGHT_SPECTRUM'
    init_args["flag_column"] = 'FLAG'
    init_args["gain_table"] = None
    init_args["max_field_of_view"] = fov
    init_args["overwrite"] = True
    init_args["channels_per_image"] = 1
    _init(**init_args)

    # grid data to produce dirty image
    grid_args = {}
    for key in schema.grid["inputs"].keys():
        grid_args[key] = schema.grid["inputs"][key]["default"]
    # overwrite defaults
    grid_args["output_filename"] = outname
    grid_args["nband"] = nchan
    grid_args["field_of_view"] = fov
    grid_args["fits_mfs"] = False
    grid_args["psf"] = True
    grid_args["residual"] = False
    grid_args["nthreads_dask"] = 1
    grid_args["nvthreads"] = 8
    grid_args["overwrite"] = True
    grid_args["robustness"] = 0.0
    grid_args["wstack"] = True
    _grid(**grid_args)

    # run clean
    spotless_args = {}
    for key in schema.spotless["inputs"].keys():
        spotless_args[key] = schema.spotless["inputs"][key]["default"]
    spotless_args["output_filename"] = outname
    spotless_args["nband"] = nchan
    spotless_args["niter"] = 2
    tol = 1e-5
    spotless_args["tol"] = tol
    spotless_args["gamma"] = 1.0
    spotless_args["pd_tol"] = 5e-4
    spotless_args["rmsfactor"] = 0.1
    spotless_args["l1reweight_from"] = 5
    spotless_args["bases"] = 'self,db1,db2,db3'
    spotless_args["nlevels"] = 3
    spotless_args["nthreads_dask"] = 1
    spotless_args["nvthreads"] = 8
    spotless_args["scheduler"] = 'sync'
    spotless_args["wstack"] = True
    spotless_args["epsilon"] = epsilon
    spotless_args["fits_mfs"] = False
    _spotless(**spotless_args)


    # get the inferred model
    dds = xds_from_zarr(f'{outname}_I_main.dds')
    freqs_dds = []
    times_dds = []
    for ds in dds:
        freqs_dds.append(ds.freq_out)
        times_dds.append(ds.time_out)

    freqs_dds = np.array(freqs_dds)
    times_dds = np.array(times_dds)
    freqs_dds = np.unique(freqs_dds)
    times_dds = np.unique(times_dds)
    ntime_dds = times_dds.size
    nfreq_dds = freqs_dds.size

    model_inferred = np.zeros((ntime_dds, nfreq_dds, nx, ny))
    for ds in dds:
        model_inferred[ds.timeid, ds.bandid, :, :] = ds.MODEL.values

    model2comps_args = {}
    for key in schema.model2comps["inputs"].keys():
        model2comps_args[key] = schema.model2comps["inputs"][key]["default"]
    model2comps_args["output_filename"] = outname
    model2comps_args["nbasisf"] = nchan
    model2comps_args["fit_mode"] = 'Legendre'
    model2comps_args["overwrite"] = True
    model2comps_args["use_wsum"] = False
    model2comps_args["sigmasq"] = 1e-14
    _model2comps(**model2comps_args)

    mds_name = f'{outname}_I_main_model.mds'
    mds = xds_from_zarr(mds_name)[0]

    # grid spec
    cell_rad = mds.cell_rad_x
    cell_deg = np.rad2deg(cell_rad)
    nx = mds.npix_x
    ny = mds.npix_y
    x0 = mds.center_x
    y0 = mds.center_y
    radec = (mds.ra, mds.dec)

    # model func
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(mds.params.values))
    symexpr = parse_expr(mds.parametrisation)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(mds.texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(mds.fexpr)
    ffunc = lambdify(params[1], fexpr)

    # model coeffs
    coeffs = mds.coefficients.values
    locx = mds.location_x.values
    locy = mds.location_y.values

    model_test = np.zeros((ntime_dds, nfreq_dds, nx, ny), dtype=float)
    for i in range(ntime_dds):
        tout = tfunc(times_dds[i])
        for j in range(nchan):
            fout = ffunc(freqs_dds[j])
            model_test[i,j,locx,locy] = modelf(tout, fout, *coeffs)

    # models need to match exactly
    assert_allclose(1 + model_test, 1 + model_inferred)

    # degrid from coeffs
    degrid_args = {}
    for key in schema.degrid["inputs"].keys():
        degrid_args[key] = schema.degrid["inputs"][key]["default"]
    # overwrite defaults
    degrid_args["ms"] = str(test_dir / 'test_ascii_1h60.0s.MS')
    degrid_args["mds"] = f'{outname}_I_main_model.mds'
    degrid_args["channels_per_image"] = 1
    degrid_args["nthreads_dask"] = 1
    degrid_args["nvthreads"] = 8
    degrid_args["wstack"] = True
    _degrid(**degrid_args)

    # manually place residual in CORRECTED_DATA
    resid = xds.DATA.data - xds.MODEL_DATA.data
    xds['DATA'] = (('row','chan','coor'), resid)
    writes = [xds_to_table(xds, ms_name, columns='DATA')]
    dask.compute(writes)

    # gridding CORRECTED_DATA should return identical residuals
    # overwrite defaults
    grid_args['DATA_COLUMN'] = 'CORRECTED_DATA'
    grid_args["output_filename"] = outname
    grid_args["postfix"] = 'resid'
    grid_args["nband"] = nchan
    grid_args["field_of_view"] = fov
    grid_args["fits_mfs"] = False
    grid_args["psf"] = False
    grid_args["weight"] = False
    grid_args["residual"] = False
    grid_args["nthreads_dask"] = 1
    grid_args["nvthreads"] = 8
    grid_args["overwrite"] = True
    grid_args["robustness"] = 0.0
    grid_args["wstack"] = True
    _grid(**grid_args)

    ddso = xds_from_zarr(f'{outname}_I_resid.dds')
    for dso, ds in zip(ddso, dds):
        assert_allclose(1 + np.abs(dso.DIRTY.values),
                        1 + np.abs(ds.DIRTY.values))

# test_spotless()



import numpy as np
from functools import partial
import dask.array as da
from numpy.testing import assert_array_almost_equal
from pfb.prox.prox_21 import prox_21
from pfb.prox.prox_21m import prox_21m
from pfb.prox.prox_21m import prox_21m_numba, dual_update, dual_update_numba
from pfb.operators.psi import im2coef
from pfb.operators.psi import coef2im
import pywt
import pytest
from numba.typed import Dict
from pfb.wavelets.wavelets import (wavedecn, waverecn, unravel_coeffs,
                                   ravel_coeffs, wavelet_setup)

pmp = pytest.mark.parametrize

@pmp("nx", [128, 250])
@pmp("ny", [64, 78])
@pmp("nband", [1, 3, 6])
@pmp("nlevels", [1, 2])
def test_psi(nx, ny, nband, nlevels):
    """
    Check that decomposition + reconstruction is the identity
    """
    # image = pywt.data.aero()
    image = np.random.randn(nx, ny)
    nu = 1.0  + 0.1 * np.arange(nband)

    x = image[None, 0:nx, 0:ny] * nu[:, None, None] ** (-0.7)

    # set up dictionary
    bases = ('self','db1','db2','db3','db4','db5')
    nbasis = len(bases)
    iys, sys, ntot, nmax = wavelet_setup(x[0:1], bases, nlevels)
    ntot = tuple(ntot)

    psiH = partial(im2coef, bases=bases, ntot=ntot, nmax=nmax,
                   nlevels=nlevels)
    psi = partial(coef2im, bases=bases, ntot=ntot,
                  iy=iys, sy=sys, nx=nx, ny=ny)

    # make sure this works even when output arrays are randomly populated
    alpha = np.random.randn(nband, nbasis, nmax)  #, dtype=x.dtype)
    xrec = np.random.randn(nband, nx, ny)  #, dtype=x.dtype)

    # decompose
    psiH(x, alpha)
    # reconstruct
    psi(alpha, xrec)

    # the nbasis is required here because the operator is not normalised
    assert_array_almost_equal(nbasis*x, xrec, decimal=12)

@pmp("nx", [120, 240])
@pmp("ny", [32, 150])
@pmp("nband", [1, 3, 6])
@pmp("nlevels", [1, 2])
def test_prox21(nx, ny, nband, nlevels):
    """
    Check that applying the prox with zero step size is the identity
    """
    image = pywt.data.aero()

    nu = 1.0  + 0.1 * np.arange(nband)

    x = image[None, 0:nx, 0:ny] * nu[:, None, None] ** (-0.7)

    # set up dictionary info
    bases = ('self','db1','db2','db3','db4','db5')
    nbasis = len(bases)
    iys, sys, ntot, nmax = wavelet_setup(x, bases, nlevels)
    ntot = tuple(ntot)
    psiH = partial(im2coef, bases=bases, ntot=ntot, nmax=nmax,
                   nlevels=nlevels)
    psi = partial(coef2im, bases=bases, ntot=ntot,
                  iy=iys, sy=sys, nx=nx, ny=ny)

    weights_21 = np.random.random(nbasis*nmax).reshape(nbasis, nmax)
    sig_21 = 0.0

    alpha = np.zeros((nband, nbasis, nmax), dtype=x.dtype)
    xrec = np.zeros((nband, nx, ny), dtype=x.dtype)

    psiH(x, alpha)

    y = prox_21(alpha, sig_21, weights_21)

    psi(y, xrec)

    # the nbasis is required here because the operator is not normalised
    assert_array_almost_equal(nbasis*x, xrec, decimal=12)


@pmp("nx", [1202, 240])
@pmp("ny", [324, 1506])
@pmp("nband", [1, 3, 6])
@pmp("nlevels", [1, 2])
def test_prox21m(nx, ny, nband, nlevels):
    """
    Check that applying the prox with zero step size is the identity
    """
    image = np.random.randn(nx, ny)

    nu = 1.0  + 0.1 * np.arange(nband)

    x = image[None, :, :] * nu[:, None, None] ** (-0.7)

    # set up dictionary info
    bases = ('self','db1','db2','db3','db4','db5')
    nbasis = len(bases)
    iys, sys, ntot, nmax = wavelet_setup(x, bases, nlevels)
    ntot = tuple(ntot)
    psiH = partial(im2coef, bases=bases, ntot=ntot, nmax=nmax,
                   nlevels=nlevels)
    psi = partial(coef2im, bases=bases, ntot=ntot,
                  iy=iys, sy=sys, nx=nx, ny=ny)

    weights_21 = np.random.random(nbasis*nmax).reshape(nbasis, nmax)
    sig_21 = 0.0

    alpha = np.zeros((nband, nbasis, nmax), dtype=x.dtype)
    xrec = np.zeros((nband, nx, ny), dtype=x.dtype)

    psiH(x, alpha)

    y = prox_21m(alpha, sig_21, weight=weights_21)

    psi(y, xrec)

    # the nbasis is required here because the operator is not normalised
    assert_array_almost_equal(nbasis*x, xrec, decimal=12)


@pmp("nmax", [1234, 240, 8765])
@pmp("nbasis", [1, 5])
@pmp("nband", [1, 3, 6])
@pmp("lam", [1.0, 1e-1, 1e-3])
@pmp("sigma", [75.0, 1.0, 1e-3])
def test_prox21m_numba(nband, nbasis, nmax, lam, sigma):
    # check numba implementation matches numpy even when output contains random
    # numbers initially
    v = np.random.randn(nband, nbasis, nmax)
    vout = np.random.randn(nband, nbasis, nmax)
    l1weight = np.random.random(nbasis*nmax).reshape(nbasis, nmax)
    sigma = 1e-3
    res = prox_21m(v, lam, weight=l1weight)
    prox_21m_numba(v, vout, lam, weight=l1weight)

    assert_array_almost_equal(res, vout, decimal=12)

    res = prox_21m(v/sigma, lam/sigma, weight=l1weight)
    prox_21m_numba(v, vout, lam, sigma=sigma, weight=l1weight)

    assert_array_almost_equal(res, vout, decimal=8)


@pmp("nx", [1202, 240])
@pmp("ny", [324, 1506])
@pmp("nband", [1, 3, 6])
@pmp("nlevels", [1, 2])
def test_dual_update(nx, ny, nband, nlevels):
    """
    Compare numpy to numba optimised dual update
    """
    image = np.random.randn(nx, ny)

    nu = 1.0  + 0.1 * np.arange(nband)

    x = image[None, :, :] * nu[:, None, None] ** (-0.7)

    xp = x.copy()

    # set up dictionary info
    bases = ('self','db1','db2','db3','db4','db5')
    nbasis = len(bases)
    iys, sys, ntot, nmax = wavelet_setup(x, bases, nlevels)
    ntot = tuple(ntot)
    psiH = partial(im2coef, bases=bases, ntot=ntot, nmax=nmax,
                   nlevels=nlevels)
    psi = partial(coef2im, bases=bases, ntot=ntot,
                  iy=iys, sy=sys, nx=nx, ny=ny)

    weight21 = np.random.random(nbasis*nmax).reshape(nbasis, nmax)
    lam21 = 0.1
    sigma = 1.75

    v = np.random.randn(nband, nbasis, nmax)
    vp = v.copy()
    res1 = dual_update(v, x, psiH, lam21, sigma=sigma, weight=weight21)

    # initialise v with psiH(x)
    psiH(xp, v)
    dual_update_numba(vp, v, lam21, sigma=sigma, weight=weight21)
    assert_array_almost_equal(1 + res1,1 + v, decimal=9)

# test_dual_update(1024, 2056, 3, 2)



import numpy as np
from numpy.testing import assert_array_almost_equal
from pfb.utils.misc import kron_matvec
from africanus.gps.kernels import exponential_squared as expsq
import pytest
pmp = pytest.mark.parametrize

@pmp("nx", [120, 240])
@pmp("ny", [32, 150])
@pmp("nband", [2, 6])
@pmp("sigma0", [0.1, 1.0])
@pmp("length_scale", [1.0, 1.5])
def test_matvec(nx, ny, nband, sigma0, length_scale):
    v = np.arange(-(nband//2), nband//2)
    x = np.arange(-(nx//2), nx//2)
    y = np.arange(-(ny//2), ny//2)

    A1 = expsq(v, v, sigma0, length_scale)
    A2 = expsq(x, x, 1.0, length_scale)
    A3 = expsq(y, y, 1.0, length_scale)

    sigma = 1e-13
    C1 = np.linalg.pinv(A1, hermitian=True, rcond=sigma)
    C2 = np.linalg.pinv(A2, hermitian=True, rcond=sigma)
    C3 = np.linalg.pinv(A3, hermitian=True, rcond=sigma)

    A = (A1, A2, A3)
    C = (C1, C2, C3)

    xi = np.random.randn(nband*nx*ny)
    res = kron_matvec(A, xi)

    rec = kron_matvec(C, res)

    assert_array_almost_equal(rec, xi, decimal=5)


# flake8: noqa
import click


@click.group()
def cli():
    pass


from pfb.workers import (init, grid, degrid,
                         clean, restore, fwdbwd,
                         spotless, model2comps,
                         forward)



import numpy as np
import numba
import numexpr as ne
import concurrent.futures as cf
import pywt
import dask.array as da
from pfb.wavelets.wavelets import wavedecn, waverecn, ravel_coeffs, unravel_coeffs


@numba.njit(nogil=True, fastmath=True, cache=True)
def pad(x, n):
    '''
    pad 1D array by n zeros
    '''
    return np.append(x, np.zeros(n, dtype=x.dtype))


def _coef2im_impl(a, base, l, iy, sy, nx, ny):
    if base == 'self':
        wave = a.reshape(nx, ny)
    else:
        alpha_rec = unravel_coeffs(
            a, iy, sy, output_format='wavedecn')
        wave = waverecn(alpha_rec, base, mode='zero')
    return wave, l


def coef2im(alpha, x, bases, ntot, iy, sy, nx, ny, nthreads=1):
    '''
    Per band coefficients to image
    '''
    nband, nbasis, nmax = alpha.shape
    futures = []
    x[...] = 0.0
    with cf.ThreadPoolExecutor(max_workers=nthreads) as executor:
        for l in range(nband):
            for b in range(nbasis):
                base = bases[b]
                a = alpha[l, b, 0:ntot[b]]
                if base == 'self':
                    iyt = 1.0
                    syt = 1.0
                else:
                    iyt = iy[base]
                    syt = sy[base]
                fut = executor.submit(_coef2im_impl, a, base, l, iyt, syt, nx, ny)
                futures.append(fut)

        for f in cf.as_completed(futures):
            wave, l = f.result()
            ne.evaluate('x + wave', local_dict={
                        'x': x[l],
                        'wave': wave},
                        out=x[l], casting='same_kind')


def _im2coef_impl(x, base, l, b, nlevels):
    if base == 'self':
        wave = x.ravel()  #, mode='constant')
    else:
        wave = wavedecn(x, base, mode='zero', level=nlevels)
        wave, _, _ = ravel_coeffs(wave)
    return wave, l, b


def im2coef(x, alpha, bases, ntot, nmax, nlevels, nthreads=1):
    '''
    Per band image to coefficients
    '''
    nbasis = len(bases)
    nband, _, _ = x.shape
    alpha[...] = 0.0
    futures = []
    with cf.ThreadPoolExecutor(max_workers=nthreads) as executor:
        for l in range(nband):
            for b in range(nbasis):
                base = bases[b]
                fut = executor.submit(_im2coef_impl, x[l], base, l, b, nlevels)
                futures.append(fut)

        for f in cf.as_completed(futures):
            wave, l, b = f.result()
            alpha[l, b, 0:ntot[b]] = wave


# TODO - compare threadpool with dask versions
# def _coef2im(alpha, bases, ntot, iy, sy, nx, ny):
#     return _coef2im_impl(alpha[0][0], bases, ntot,
#                          iy, sy, nx, ny)

# def coef2im(alpha, bases, ntot, iy, sy, nx, ny, compute=True):
#     if not isinstance(alpha, da.Array):
#         alpha = da.from_array(alpha, chunks=(1, -1, -1), name=False)

#     graph = da.blockwise(_coef2im, ("band", "nx", "ny"),
#                          alpha, ("band", "basis", "ntot"),
#                          bases, None, #("basis",),
#                          ntot, None, #("basis",),
#                          iy, None,
#                          sy, None,
#                          nx, None,
#                          ny, None,
#                          new_axes={'nx': nx, 'ny': ny},
#                          dtype=alpha.dtype,
#                          align_arrays=False)
#     if compute:
#         return graph.compute()
#     else:
#         return graph


# def _im2coef(x, bases, ntot, nmax, nlevels):
#     return _im2coef_impl(x[0][0], bases, ntot, nmax, nlevels)


# def im2coef(x, bases, ntot, nmax, nlevels, compute=True):
#     if not isinstance(x, da.Array):
#         x = da.from_array(x, chunks=(1, -1, -1), name=False)

#     graph = da.blockwise(_im2coef, ("band", "basis", "nmax"),
#                          x, ("band", "nx", "ny"),
#                          bases, None, #("basis",),
#                          ntot, None, #("basis",),
#                          nmax, None,
#                          nlevels, None,
#                          new_axes={'basis':len(bases), 'nmax':nmax},
#                          dtype=x.dtype)

#     if compute:
#         return graph.compute()
#     else:
#         return graph


@numba.njit(nogil=True, fastmath=True, cache=True, parallel=True)
def im2coef_dist(x, bases, ntot, nmax, nlevels):
    '''
    Per band image to coefficients
    '''
    nbasis = len(bases)
    alpha = np.zeros((nbasis, nmax), dtype=x.dtype)
    for b in numba.prange(nbasis):
        base = bases[b]
        if base == 'self':
            # ravel and pad
            alpha[b] = pad(x.ravel(), nmax-ntot[b]) #, mode='constant')
        else:
            continue
            # # decompose
            # alpha_tmp = wavedecn(x, base, mode='zero', level=nlevels)
            # # ravel and pad
            # alpha_tmp, _, _ = ravel_coeffs(alpha_tmp)
            # alpha[b] = pad(alpha_tmp, nmax-ntot[b])  #, mode='constant')

    return alpha


@numba.njit(nogil=True, fastmath=True, cache=True, parallel=True)
def coef2im_dist(alpha, bases, ntot, iy, sy, nx, ny):
    '''
    Per band coefficients to image
    '''
    nbasis = len(bases)
    x = np.zeros((nbasis, nx, ny), dtype=alpha.dtype)
    for b in numba.prange(nbasis):
        base = bases[b]
        a = alpha[b, 0:ntot[b]]
        if base == 'self':
            wave = a.reshape(nx, ny)
        else:
            continue
            # alpha_rec = unravel_coeffs(
            #     a, iy[base], sy[base], output_format='wavedecn')
            # wave = waverecn(alpha_rec, base, mode='zero')

        x[b] = wave
    return np.sum(x, axis=0)

